"""
======================================
  ViPErLEED Graphical User Interface
======================================
    *** module guilib.leedsim ***

Module: mainwindow.py

Created: 2020-01-13
Author: Michele Riva

This is the main window of the ViPErLEED simulation plug-in. This window
handles all the widgets and all the calls to other dialogs.

2020-01-13 At present this is the only plug-in available; the behavior and
           coding of this will change when other plug-ins are integrated
"""

#TESTING
from time import perf_counter
#TESTING

import re
import copy

import numpy as np
import PyQt5.QtCore as qtc
import PyQt5.QtWidgets as qtw

from viperleed import guilib as gl


def show_pre_release_popup():
    """
    Show a pop-up dialog informing that the current version is a not meant
    to be distributed without permission, as it is a pre-release
    """
    txt = (f"ViPErLEED v{gl.GLOBALS['version']} is a pre-release not meant for "
           "public distribution!<p> Contact us at <a href="
           u''"'mailto:riva@iap.tuwien.ac.at'"'>riva@iap.tuwien.ac.at</a> </p>')
    msgBox = qtw.QMessageBox(qtw.QMessageBox.Information, "Pre-release notice",
                             txt)
    msgBox.exec()


def show_use_betatest_version_popup():
    """Show a pop-up dialog hinting at using the betatest version."""
    _betatest = 'https://github.com/viperleed/viperleed-betatest'
    txt = (
        f'The ViPErLEED graphical user interface for v{gl.GLOBALS["version"]} '
        'is currently under development.<p>'
        'Please use the pre-packed version available from the <a href='
        f'{_betatest}/releases/latest>viperleed-betatest'
        '</a> GitHub repository (file gui.zip).</p>'
        )
    msgBox = qtw.QMessageBox(qtw.QMessageBox.Information,
                             'Use betatest version',
                             txt)
    msgBox.exec_()
    

@gl.broadcast_mouse
class LEED_GUI(qtw.QMainWindow):
    pre_release = False
    extension = '*.tlm'
    version = gl.GLOBALS['version']
    
    extStr = 'LEED input files ({})'.format(' '.join([extension,
                                                      extension.upper()]))
    extStr = ';;'.join([extStr, 'All files (*)'])
    default_open = ('./guilib/leedsim/input examples/', 'PatternInfo.tlm')
    default_export = ('./guilib/leedsim/exported/', 'LEEDSpots.csv')
    
    # these are the parameters to read from the LEED file
    inputParams = ('eMax', 'surfBasis', 'SUPERLATTICE', 'surfGroup',
                   'bulkGroup')
    optionalParams = ('bulk3Dsym', 'screenAperture')
    
    def __init__(self):
        super().__init__()
        self.fonts = gl.AllGUIFonts()
        self.leedParams = dict()
        self.openedFile = None  # Keep track of the name of the currently 
                                # open file. This is used for exporting.
        self.initUI()
        self.centerOnScreen()
        if self.pre_release:
            show_pre_release_popup()
    
    def centerOnScreen(self):
        qr = self.frameGeometry()
        cp = qtw.QDesktopWidget().availableGeometry(self).center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())
    
    def setMenu(self):
        menubar = self.menuBar()
        
        self.actionsToEnable = []
        
        # "File"
        self.fileMenu = menubar.addMenu('&File')
        
        # File -> New
        newF = qtw.QAction('&New...', self)
        newF.setShortcut('Ctrl+N')
        newF.setStatusTip('New LEED pattern file')
        newF.triggered.connect(self.fileNewDialog)
        
        # File -> Open
        openF = qtw.QAction('&Open...', self)
        openF.setShortcut('Ctrl+O')
        openF.setStatusTip('Open LEED pattern file')
        openF.triggered.connect(self.fileOpenDialog)
        
        # File -> Save
        saveF = qtw.QAction('&Save', self)
        saveF.setShortcut('Ctrl+S')
        saveF.setStatusTip('Save LEED pattern file')
        saveF.triggered.connect(self.fileSavePressed)
        
        self.actionsToEnable.append(saveF)
        
        # File -> Save As...
        svAs = qtw.QAction('&Save As...', self)
        svAs.setShortcut('Ctrl+Shift+S')
        svAs.setStatusTip('Save LEED pattern file as...')
        svAs.triggered.connect(self.fileSaveAsPressed)
        self.actionsToEnable.append(svAs)
        
        # File -> Export sub-menu...
        exportMenu = qtw.QMenu('Export')
        self.actionsToEnable.append(exportMenu)
        
        # ...and its contents:
        # * Export list of beams to *.csv
        exportCSV = qtw.QAction('&Export Beam List...', self)
        exportCSV.setShortcut('Ctrl+E')
        exportCSV.setStatusTip('Export list of LEED beams to *.csv')
        exportCSV.triggered.connect(self.exportBeamsPressed)
        self.actionsToEnable.append(exportCSV)
        
        exportMenu.addAction(exportCSV)
        
        # File -> Exit
        exit = qtw.QAction('&Exit', self)
        exit.setShortcut('Ctrl+Q')
        exit.setStatusTip('Exit application')
        exit.triggered.connect(qtw.qApp.quit)
        #exitAct.triggered.connect(self.close)
        # Replacing line 109 with 110 makes the popup appear also when
        # closing with CTRL+Q 
        
        self.fileMenu.addAction(newF)
        self.fileMenu.addAction(openF)
        self.fileMenu.addAction(saveF)
        self.fileMenu.addAction(svAs)
        self.fileMenu.addMenu(exportMenu)
        self.fileMenu.addSeparator()
        self.fileMenu.addAction(exit)
    
    def setToolBar(self):
        self.tBar = self.addToolBar('')
        self.tBar.setFloatable(False)
        self.tBar.setIconSize(0.7*self.tBar.iconSize())
        
        newF = qtw.QAction(self.style().standardIcon(qtw.QStyle.SP_FileIcon),
                       'New LEED pattern', self)
        newF.setStatusTip('New LEED pattern file (Ctrl+N)')
        newF.triggered.connect(self.fileNewDialog)
        
        openF = qtw.QAction(
            self.style().standardIcon(qtw.QStyle.SP_DialogOpenButton),
            'Open LEED pattern', self)
        openF.setStatusTip('Open LEED pattern file (Ctrl+O)')
        openF.triggered.connect(self.fileOpenDialog)
        
        saveF = qtw.QAction(
            self.style().standardIcon(qtw.QStyle.SP_DialogSaveButton),
            'Save LEED pattern', self)
        saveF.setStatusTip('Save LEED pattern file (Ctrl+S)')
        saveF.triggered.connect(self.fileSavePressed)
        
        exportAct = qtw.QAction(
            self.style().standardIcon(qtw.QStyle.SP_DriveFDIcon),
            'Export LEED pattern', self)
        exportAct.setStatusTip('Export list of LEED beams to '
                               'comma-separated file (Ctrl+E)')
        exportAct.triggered.connect(self.exportBeamsPressed)
        
        self.tBar.addAction(newF)
        self.tBar.addAction(openF)
        self.tBar.addAction(saveF)
        self.tBar.addAction(exportAct)
        
        self.actionsToEnable.append(saveF)
        self.actionsToEnable.append(exportAct)
    
    def fileOpenDialog(self):
        fname = qtw.QFileDialog.getOpenFileName(self, 'Open LEED pattern',
                                                self.default_open[0],
                                                self.extStr)
        loadedParams = self.parseLEEDFile(fname)
        if loadedParams is not None:
            self.openedFile = fname
            self.leedParams = loadedParams
            self.loadLEEDInput(loadedParams, True)
            self.isSaved = True
            self.openFile = fname[0]
        else:
            self.fileReadError(fname=fname[0])
    
    def fileNewDialog(self):
        if not hasattr(self, 'newDialog'):
            self.newDialog = gl.NewFileDialog(self, self.leedParams)
            self.newDialog.dialogEdited.connect(self.loadLEEDInput)
            self.isSaved = False
        else:
            self.newDialog.open(self.leedParams)
    
    def fileSavePressed(self):
        if (not hasattr(self, 'openFile')) or (self.openFile is None):
            # no file is currently open
            self.fileSaveAsPressed()
        else:
            # TODO: ADD HERE A MESSAGE BOX ASKING IF YOU WANT TO OVERWRITE?
            self.saveToFile(self.openFile)
            self.isSaved = True
    
    def fileSaveAsPressed(self):
        fname = qtw.QFileDialog.getSaveFileName(self, 'Save TLM input', '',
                                                self.extStr)
        if fname is not None and fname[0]:
            # file selected correctly
            self.saveToFile(fname[0])
            self.openFile = fname[0]
            self.isSaved = True
    
    def saveToFile(self, fname):
        # Sort out the parameters, converting the values to strings
        params = copy.deepcopy(self.leedParams)
        for key, val in params.copy().items():  # BUG: dictionary changes size upon saving
            if key == 'SUPERLATTICE' or key == 'surfBasis':
                params[key] = np.array2string(
                                        val,
                                        separator=',',
                                       # precision=5,
                                        suppress_small=True
                                        ).replace('\n', '').replace(' ', '')
            elif key in ('eMax', 'screenAperture'):
                params[key] = str(val)
            elif key in ('bulkGroup', 'surfGroup'):
                params[key] = val.group
            elif key == 'bulk3Dsym' and val is None:
                del params[key]
        # now params contains correctly formatted strings
        params = list(params.items())
        paramsTxt = '\n'.join('{}'.format('='.join(param))
                              for param in params)
        with open(fname, 'w+') as f:
            f.write(paramsTxt)
    
    def exportBeamsPressed(self):
        # Create a dialog to select which domains should be exported
        # and to process the data to export
        if not hasattr(self, 'exportDialog'):
            self.exportDialog = gl.ExportCSVDialog(self.leed,
                                                   parent=self)
            self.exportDialog.exportSelected.connect(self.exportCSV)
        else:
            self.exportDialog.open()
    
    def parseLEEDFile(self, fname, checkOnly=False):
        # returns: None if file is unreadable/no file was selected; empty
        # dict if file does not contain all parameters/incorrect ones; dict
        # with parameters otherwise
        self.fileErr = ''   # keeps track of what happened
        
        if fname[0]:
            with open(fname[0], 'r') as f:
                try:
                    flines = f.readlines()
                except:
                    ext = qtc.QFileInfo(fname).suffix().lower()
                    self.fileErr = 'ERROR: File type *.%s not supported.'%(ext)
                    return None
        else:
            self.fileErr = 'No file selected.'
            return None
        
        if checkOnly:
            return dict()
        
        if flines is None:
            return dict()
        
        # Something has been successfully read
        # get rid of white spaces (i.e, ' ', '\t', '\n', '\r')
        flines = [re.sub(r'\s+', '', line) for line in flines]
        
        #get rid of comments, that begin with \# \! \%
        commentChars = ['#', '!', '%']
        for ch in commentChars:
            flines = [line.partition(ch)[0].lower() for line in flines]
            # i.e., take whatever block is on the left side of any of the
            # comment characters
        
        # now keep only lines that contain exactly one '=' sign
        flines = [line for line in flines if line.count('=') == 1]
        
        if len(flines) == 0:
            self.fileErr = 'File does not contain any LEED parameter.'
            return dict()

        # and split them in two:
        # _param contains what's on the left of '=',
        # _values what's on the right
        flines_param, flines_values = zip(
                         *[np.array(line.partition('='))[[0, 2]].tolist()
                           for line in flines
                           if len(line) > 1]
                         )
        # make a dictionary out of these. This also removes duplicate
        # parameters, taking the last occurrence as the value
        par_dict = dict(zip(flines_param, flines_values))
        
        if len(par_dict) < len(flines_param):
            self.fileErr = ('Warning: Repeated entries found;'
                            ' using last occurrence.')
        
        # remove all entries that are not acceptable parameters
        acceptable = [p.lower()
                      for p in (*self.inputParams, *self.optionalParams)]
        par_dict = {key.lower(): val
                    for key, val in par_dict.items()
                    if key.lower() in acceptable}

        # replace the dictionary keys in par_dict with their standard
        # capitalization, found in self.inputParams and in self.optionalParams.
        # Will make a difference between keys that are mandatory parameters
        # and keys that are optional ones.
        for par in (*self.inputParams, *self.optionalParams):
            try:
                par_dict[par.lower()]
            except KeyError:
                if par in self.inputParams:
                    self.fileErr = (f'ERROR: Mandatory parameter {par} '
                                    'not found. '
                                    + self.fileErr)
                    return dict()
                # otherwise the parameter is optional, and will be set to a
                # default of 'None'
                par_dict[par] = 'None'
            else:
                par_dict[par] = par_dict.pop(par.lower())

        par_dict['eMax'] = float(par_dict['eMax'])
        par_dict['surfBasis'] = gl.string_matrix_to_numpy(
            par_dict['surfBasis'], dtype=float, needs_shape=(2, 2)
            )
        par_dict['SUPERLATTICE'] = gl.string_matrix_to_numpy(
            par_dict['SUPERLATTICE'], int, needs_shape=(2, 2)
            )
        
        try:
            par_dict['screenAperture'] = float(par_dict['screenAperture'])
        except ValueError:
            self.fileErr += ("Warning: screenAperture not found, or value "
                             " not acceptable. Will use default 110 deg.")
            par_dict['screenAperture'] = 110.0
        
        # Now handle the exit conditions
        try:
            b_group = gl.PlaneGroup(par_dict['bulkGroup'])
        except ValueError:
            self.fileErr = ('ERROR: Unknown bulk group type '
                           + par_dict['bulkGroup']
                           + self.fileErr)
            return dict()
        else:
            par_dict['bulkGroup'] = b_group
        
        try:
            s_group = gl.PlaneGroup(par_dict['surfGroup'])
        except ValueError:
            self.fileErr = ('ERROR: Unknown surface group type '
                            + par_dict['surfGroup']
                            + self.fileErr)
            return dict()
        else:
            par_dict['surfGroup'] = s_group
        
        if ((par_dict['surfBasis'] is None)
            or (par_dict['SUPERLATTICE'] is None)):
            self.fileErr  = ('ERROR: Too many entries in '
                            + 'surface unit vectors or SUPERLATTICE. '
                            + self.fileErr)
            return dict()
        return par_dict
    
    def exportCSV(self, params):
        # Ask if the user wants to save the input if it isn't
        if not self.isSaved:
            self.unsavedPopup()
        
        # Then open a normal file dialog to save the data to file
        fname = qtw.QFileDialog.getSaveFileName(self, 'Export list of beams',
                                                self.default_export[0],
                                                'Comma-separated file (*.csv)')
        if not fname[0]:
            return
        
        # set up the other parameters needed for export_pattern_csv
        if hasattr(self, 'openFile') and self.openFile:
            params['source'] = self.openFile
        params['version'] = self.version
        
        gl.export_pattern_csv(fname[0], (self.leed,), **params)
    
    def unsavedPopup(self):
        reply = qtw.QMessageBox.question(self, 'Edits unsaved',
                                        'Would you like to save changes?',
                                        qtw.QMessageBox.Yes
                                        | qtw.QMessageBox.No,
                                        qtw.QMessageBox.Yes)
        if reply == qtw.QMessageBox.Yes:
            self.fileSavePressed()
    
    def loadLEEDInput(self, par_dict, readFile=False):
        if par_dict:
            self.leedParams = par_dict
            self.set_ControlsActive(True)
            self.initRealAndLEED(True)
            self.enableActions(True)
        if readFile:
            self.fileReadError()
        if hasattr(self, 'exportDialog'):
            # delete exportDialog as the number of its widgets (that depends
            # on the number of domains) is most likely not accurate anymore.
            self.exportDialog.destroy()
            del self.exportDialog

    def insertText(self, label, text, align='left'):
        # sets, reshapes, and realigns horizontally a QWidget after changing
        # its text
        
        if align not in ['left', 'center', 'right']:
            raise ValueError("Alignment of QWidget can only be 'left',"
                             "'center' or 'right'")
        
        pos0 = label.geometry().topLeft()
        if align == 'left':
            delta0 = 0
        elif align == 'right':
            delta0 = label.width()
        else:
            delta0 = label.width()/2
        
        label.setText(text)
        label.adjustSize()
        
        if align == 'left':
            delta1 = 0
        elif align == 'right':
            delta1 = label.width()
        else:
            delta1 = label.width()/2
        
        label.move(pos0.x() - delta1 + delta0, pos0.y())
    
    def initControls(self):
        # Initialize all controls and labels, and set their dimensions
        # w is the centralWidget
        
        w = self.window().centralWidget()
        
        #### REAL SPACE AND LEED PATTERN PLOTTING CANVASES
        self.realSpace = gl.RealCanvas(parent=w, title='Real Space Lattice')
        self.recSpace = gl.LEEDCanvas(parent=w, title='LEED Pattern')
            
        # draw the LEED screen as a circle. It will also act as a background,
        # and will be used as a clip path for the pattern
        self.recSpace.initLEEDScreen(
                               lineThick=self.realSpace.getSpinesThickness())
        
        #### ROTATION ####
        self.rotWidg = gl.RotationBlock(w)
        self.realSpace.wheel_buddy = self.rotWidg.text
        
        #### ENERGY ####
        self.enWidg = gl.EnergyBlock(w)
        self.recSpace.wheel_buddy = self.enWidg.text
        
        #### DOMAINS ####
        self.doms = gl.DomsBlock(w)
        
        #### Unit cells shape and group ####
        self.cellShapes = qtw.QLabel("Slab: \u2014, \u2014. "
                                     "Bulk: \u2014, \u2014",
                                     w)
        self.cellShapes.setFont(gl.AllGUIFonts().largeTextFont)
        self.cellShapes.setStatusTip('Open or drag-drop a LEED pattern file'
                                     'to print the shapes and symmetry'
                                     'groups of the system!')
        self.cellShapes.setAlignment(qtc.Qt.AlignHCenter | qtc.Qt.AlignBaseline)
        self.cellShapes.setSizePolicy(qtw.QSizePolicy.Expanding,
                                      qtw.QSizePolicy.Preferred)
        self.cellShapes.adjustSize()
        
        # Prepare a list of controls
        # Notice that the plot controls are at the end on purpose, 
        # so all the others are updated first after loading a file
        self.allCtrls = [self.rotWidg, *(self.rotWidg.subWidgs),
                         self.enWidg, *(self.enWidg.subWidgs),
                         self.doms, self.cellShapes,
                         self.realSpace, self.recSpace]
        
        # Disable all controls at initialization.
        # They will be enabled when a file is loaded
        self.set_ControlsActive(False)
        
    def set_ControlsActive(self, active):
        if isinstance(active, bool):
            [ctrl.setEnabled(active) for ctrl in self.allCtrls]
        else:
            raise ValueError('Invalid datatype for set_ControlsActive')

    def composeUI(self):
        # Position all controls correctly
        
        winWidg = self.window().centralWidget()
        
        g0 = winWidg.geometry()   # Origin and size of centralWidget
                                  # of the window, i.e. window size minus
                                  # size of menu and status bars
        w0 = g0.width()
        h0 = g0.height()
        self.area = w0*h0
        
        topLayout = winWidg.layout()
        '''
        Layout:
            Legend: sh/sv(n)= horizontal/vertical stretch;
                    RS= real space (VBox)
                    LD= leed (Vbox)
                    Ro= rotation block (excluding bulk buttons)
                    bu= bulk alignment buttons
                    En= energy block (excluding text below)
                    Et= energy limits text
                    Db= domains toggle button
                    Dt= domains text
                    Sh= slab shape (sh+Hbox+sh)
                    
                    if more symbols are present in a cell, the left ones correspond to widgets/items below the right ones
        
              0      1        2        3        4        5        6        7
            -------------------------------------------------------------
        0    | sv |    RS    |  sv    |  LD    |  sv    |        |        |        |    
            -------------------------------------------------------------
        1    | sv |        |  sv    |        |  sv    |        |        |        |
            -------------------------------------------------------------
        2    | sv |        |  sv    |        |  sv    |        |        |        |
            -------------------------------------------------------------
        3    | sv |        |  sv    |        |  sv    |        |        |        |
            -------------------------------------------------------------
        4    | sv |        |  sv    |        |  sv    |        |        |        |
            -------------------------------------------------------------
        5    | sv |        |  sv    |        |  sv    |        |        |        |
            -------------------------------------------------------------
        6    | sv |        |  sv    |        |  sv    |        |        |        |
            -------------------------------------------------------------
        7    | sv |        |  sv    |        |  sv    |        |        |        |
            -------------------------------------------------------------
        '''
        
        ## REAL SPACE AND LEED PATTERN
        if self.testConfig == 'layout':
            topLayout.setSpacing(3)
            topLayout.setContentsMargins(0, 0, 0, 0)
            basicStretch = 1
            plotStretch = (topLayout.columnStretch(0)
                           * self.realSpace.getStretchFactor().width())
            
            topLayout.addLayout(self.realSpace.mplLayout, 0, 1, 1, 1)
            topLayout.setColumnStretch(0, basicStretch)
            topLayout.setColumnStretch(1, plotStretch)
            topLayout.setColumnStretch(2, basicStretch)
            
            topLayout.addLayout(self.recSpace.mplLayout, 0, 3, 1, 1)
            topLayout.setColumnStretch(3, plotStretch)
            topLayout.setColumnStretch(4, basicStretch)
            
            #[topLayout.setColumnStretch(col,1) for col in [3,4,10,11]]
            #topLayout.setRowStretch(3,1)
            # topLayout.setRowStretch(0,1)
            # topLayout.setRowStretch(1,2)
            #topLayout.setRowStretch(5,1)
            #topLayout.setRowStretch(6,1)
        else:
            plotsHBox = qtw.QHBoxLayout()
            plotsHBox.addStretch(1)
            plotsHBox.addLayout(self.realSpace.mplLayout)
            plotsHBox.addStretch(1)
            plotsHBox.addLayout(self.recSpace.mplLayout)
            plotsHBox.addStretch(1)
        
            winWidg.layout().addLayout(plotsHBox)
        
        labelToTextOffset = -qtc.QPoint(0,3)

        #### ROTATION ####
        if self.testConfig == 'layout':
            pass
        else:
            rotx = (w0 - self.rotWidg.width())/2 + 0.032*w0
            self.rotWidg.move(qtc.QPoint(rotx,8))
        
        #### ENERGY ####
        if self.testConfig == 'layout':
            pass
        else:
            enx = w0*.98 - self.enWidg.width()
            self.enWidg.move(qtc.QPoint(enx,8))
        
        #### TOGGLE DOMAINS ###
        dx = self.enWidg.geometry().topRight().x() - self.doms.width()
        dy = h0*.97 - self.doms.height()
        self.doms.move(qtc.QPoint(dx,dy))
        
        #### CELLS SHAPES/GROUPS ####
        if self.testConfig == 'layout':
            self.cellShapes.setAlignment(qtc.Qt.AlignCenter)
                        #(qtc.Qt.AlignHCenter | qtc.Qt.AlignTop)
            shapesLay = qtw.QHBoxLayout()
            shapesLay.addStretch(1)
            shapesLay.addWidget(self.cellShapes,
                                qtc.Qt.AlignHCenter | qtc.Qt.AlignTop)
            shapesLay.addStretch(1)
            topLayout.addLayout(shapesLay, 1, 0, 1, topLayout.columnCount())#,qtc.Qt.AlignCenter)
            # topLayout.setRowStretch(0,100)
            # topLayout.setRowStretch(1,5)
            # vSpacer=QSpacerItem(1, 10, hPolicy = qtw.QSizePolicy.Expanding, vPolicy = qtw.QSizePolicy.Minimum)
            # topLayout.addItem(vSpacer,2,0,1,topLayout.columnCount())
            #topLayout.addWidget(self.cellShapes,1,0,1,1,qtc.Qt.AlignHCenter|qtc.Qt.AlignTop)
            #pass
        else:
            self.cellShapes.move((w0-self.cellShapes.geometry().width())/2,
                                 self.doms.geometry().center().y())
    
        ##TESTING
        # self.ensurePolished()
        # tmpscreencenter = self.recSpace.geometry().center()
        # ctmp = winWidg.mapToGlobal(self.recSpace.geometry().center())
        # print('\n\n', self.pos(), tmpscreencenter,
              # ctmp, self.recSpace.parentWidget()==winWidg, '\n')
        tmpC = np.array([500, 500])#np.array([530.5, 500.5])
        # print("\ncreating")
        self.tmps = [gl.HoverAnnot(winWidg) for i in range(12)]
        deltas = [np.array([np.cos(x), np.sin(x)]) for x in np.linspace(0, 2*np.pi, num=len(self.tmps), endpoint=False)]
        # print("assign deltas")
        for (tmp, delta) in zip(self.tmps, deltas):
            tmp.head_pos = delta*5 + tmp.center
            # tmp.show()
        
        ## END TESTING
    
    def enableActions(self, enable):
        if not isinstance(enable, bool):
            raise
        [act.setEnabled(enable) for act in self.actionsToEnable]
    
    def initUI(self):
        # print("Initializing UI...")
        # print("... Menus")
        self.setMenu()
        
        # print("... Tool bar")
        self.setToolBar()
        self.enableActions(False)
        
        self.setAcceptDrops(True)
        
        # print("... Title and icon")
        self.setWindowTitle('LEED Pattern Indexing Helper')
        self.setWindowIcon(self.style().standardIcon(qtw.QStyle.SP_DirIcon))
        
        # print("... Status bar")
        self.sBar = qtw.QStatusBar()
        self.setStatusBar(self.sBar)
        self.sBar.showMessage('Ready')
        
        self.testConfig = 'fix'#'layout'#
        cw = qtw.QWidget()
        if self.testConfig == 'layout':
            topLayout = qtw.QGridLayout(cw)
        else:
            topLayout = qtw.QHBoxLayout(cw)
        # print(f"... Central widget ({cw})")
        self.setCentralWidget(cw)
        
        self.adjustWindowSize()
        
        # print("... Initialize controls")
        self.initControls()
        # print("... Position controls")
        self.composeUI()
    
    def sizeHint(self):
        self.ensurePolished()
        screenFraction = 0.8
        screensize = qtw.QDesktopWidget().availableGeometry(self).size()
        sH = screenFraction*screensize
        return sH
    
    def adjustWindowSize(self, size=None):
        # Updates the size of the central widget, window and menus.
        if size is None:
            size = self.sizeHint()
        # the next line will need to replace the one after
        # once I make the window re-sizable
        #self.resize(size)
        self.setFixedSize(size)
        self.layout().activate() # this rebuilds the window layout, and forces the centralWidget to fill the space
    
    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.adjustWindowSize(size=event.size())
    
    def initRealAndLEED(self, active):
        if active:
            # An acceptable LEED input has been loaded.
            # Update all controls with the correct stuff
            self.real = gl.RealSpace(self.leedParams)
            self.leed = gl.LEEDPattern(self.leedParams)
            for ctrl in self.allCtrls:
                self.updateCtrlAfterFileLoad(ctrl)
            self.doms.initPopup()
            # initially set focus to a widget that does not respond to
            # wheelEvent.
            self.doms.toggle.setFocus()
        
        self.connectControlEvents(active)
        
    def connectControlEvents(self, active):
        self.buts = [self.enWidg.enUp, self.enWidg.enDown,
                     self.rotWidg.cw, self.rotWidg.ccw,
                     self.rotWidg.h10, self.rotWidg.v10, 
                     self.rotWidg.h01, self.rotWidg.v01, 
                     self.doms.toggle]
        # First disconnect all controls, as this prevents multiple connections
        # from being established every time a new file is opened
        self.disconnectAll()

        if active:
            self.enWidg.text.textModified.connect(self.energyChanged)
            self.rotWidg.text.textModified.connect(self.rotationChanged)
            [but.clicked.connect(self.buttonPressed) for but in self.buts]
    
    def disconnectAll(self):
        for (widg, slt) in zip([self.enWidg.text, self.rotWidg.text],
                               [self.energyChanged, self.rotationChanged]):
            try:
                widg.textModified.disconnect(slt)
            except TypeError:
                pass
        for but in self.buts:
            try:
                but.clicked.disconnect()
            except TypeError:
                pass
        
    def updateCtrlAfterFileLoad(self, ctrl):
        if ctrl not in self.allCtrls:
            raise ValueError('Unknown control')
        
        leed = self.leed
        real = self.real
        
        if ctrl == self.rotWidg.text:
            ctrl.setText('0.0')
        elif ctrl == self.enWidg.text:
            ctrl.setText(f'{leed.max_energy/2:.1f}')
            ctrl.setLimits(10, leed.max_energy)
        elif ctrl == self.enWidg.limits:
            ctrl.setText('Min = 10 eV\n'
                         f'Max = {leed.max_energy:.1f} eV')
        elif ctrl == self.doms:
            ctrl.updateText(f'{leed.n_domains} inequivalent domain(s)')
            ctrl.setTips(text='Click to see all the superlattice '
                              'matrices that generate the domains.')
            ctrl.hide = False
            if leed.n_domains == 1:
                ctrl.toggle.setEnabled(False)
            else:
                ctrl.toggle.setEnabled(True)
        elif ctrl == self.cellShapes:
            self.insertText(ctrl,
                            f"Slab: {real.surf.cell_shape}, {real.surf.group}. "
                            f"Bulk: {real.bulk.cell_shape}, {real.bulk.group}",
                            'center')
            ctrl.setStatusTip('Cell shapes and plane groups '
                              'for the whole slab and its bulk.')
        elif ctrl == self.realSpace:
            ctrl.plotLattices()
        elif ctrl == self.recSpace:
            ctrl.plotLEED(self.doms.hide)
        else:
            pass
            # Nothing should be done for the others
            # (except enabling and connecting events, which is done elsewhere)
        
        self.statusBar().showMessage('Ready')
    
    def get_energy(self): #is it possible to get rid of this?
        return float(self.enWidg.text.text())
    
    def get_angle(self): #is it possible to get rid of this?
        return float(self.rotWidg.text.text())
    
    def closeEvent(self, event):
        reply = qtw.QMessageBox.question(self,
                                     'Message',
                                     'Are you sure to quit?',
                                     qtw.QMessageBox.Yes | qtw.QMessageBox.No,
                                     qtw.QMessageBox.No)

        if reply == qtw.QMessageBox.Yes:
            event.accept()
        else:
            event.ignore()
    
    def mouseReleaseEvent(self, event):
        """
        Re-implement mouseReleaseEvent to hide the pop-up with matrices when
        clicking anywhere on the window
        """
        doms = self.doms
        if (doms.isEnabled()
            and doms.matricesPopup.isVisible()
            and not doms.text.underMouse()):
            # the reason for not acting when doms.text is underMouse() is that
            # the event handling in this case is done within DomsBlock
            doms.matricesPopup.hide()
        # TESTING
        # for annot in self.recSpace.annots:
            # annot.show()
        # END TESTING
        super().mouseReleaseEvent(event)
    
    # The next event handlers take care of user-induced changes 
    # on the controls
    def energyChanged(self, eOld, eNew):
        self.recSpace.plotLEED(self.doms.hide)
    
    def rotationChanged(self, rotOld, rotNew):
        self.realSpace.plotLattices()
        self.recSpace.plotLEED(self.doms.hide)
    
    def buttonPressed(self, event):
        pressed = self.sender()
        if pressed in self.enWidg.subWidgs:
            self.enWidg.on_buttonPressed(pressed)
        elif pressed in self.rotWidg.subWidgs:
            self.rotWidg.on_buttonPressed(pressed)
        elif pressed == self.doms.toggle:
            self.doms.togglePressed()
            self.recSpace.plotLEED(self.doms.hide)
    
    # The next two event handlers are for drag-drop of files
    def dragEnterEvent(self, event):
        mimeData = event.mimeData()
        if mimeData.hasUrls():   # Check that the user is dropping a file
            fname = [url.toLocalFile() for url in mimeData.urls()]
            
            # now handle differently the cases in which the user has
            # dropped multiple files:
            if len(fname) > 1:
                # for now, just take the first one. Then there will be a 
                # popup asking which one to load
                fname = fname[0]
                event.mimeData().setUrls(fname)
            # check that the file contains some readable content
            loadedParams = self.parseLEEDFile(fname, checkOnly=True)
            if loadedParams is not None:
                event.acceptProposedAction()
    
    def dropEvent(self, event):
        loadedParams = self.parseLEEDFile([url.toLocalFile()
                                           for url
                                           in event.mimeData().urls()])
        if loadedParams:
            event.acceptProposedAction()
            self.loadLEEDInput(loadedParams, True)
            self.isSaved = True
        else:
            self.fileReadError()
    
    def fileReadError(self, fname=None, msg=None):
        if msg is None:
            msg = self.fileErr
        if msg != '':
            self.statusBar().showMessage(msg)
    
    # def showEvent(self, event):
        # # print("begin showEvent")
        # super().showEvent(event)

        # # print("super().showEvent finished")
        # tmpscreencenter = self.recSpace.geometry().center()
        # tmpLayCtr = self.recSpace.mplLayout.geometry().center()
        # ctmp = self.window().centralWidget().mapToGlobal(
        # # ctmp = self.window().mapToGlobal(
               # self.recSpace.geometry().center())
        # # print('\n\n', tmpscreencenter, tmpLayCtr, ctmp, self.recSpace.parentWidget().mapToGlobal(
               # # self.recSpace.geometry().center()), self.recSpace.parentWidget().mapToGlobal(qtc.QPoint(0,0)))
        # # print('\n\n', self.pos(), tmpscreencenter,
              # # ctmp, self.recSpace.parentWidget()==winWidg, '\n')
        # for tmp in self.tmps:
            # # print(f"before: {tmp.center}", end=' ')
            # tmp.center = ctmp
            # tmp.show()
            # # print(f"after: {tmp.center}")
    
    def moveEvent(self, event):
        super().moveEvent(event)
        delta = event.pos() - event.oldPos()
        
        for annot in self.recSpace.annots:
            # NB: one cannot use += for in-place assignment since this redefines
            # the .center attribute that is a @property!
            annot.center = qtc.QPoint(*annot.center) + delta
            if annot.isVisible():
                # move only if visible to handle the situation in which the
                # window is moved while the annotations are displayed: this way
                # annotations move together with the window. If one would have
                # this call even when not visible, the move is issued before the
                # next show, and it will mess the position of the annotation
                annot.move(annot.pos() + delta)



"""POSCAR utility get_bulk_repeat.

Reads a POSCAR file, asks at what c value the bulk starts, then
automatically reduces the size of the POSCAR to non-redundant bulk
layers only, and outputs the appropriate N_BULK_LAYERS and BULK_REPEAT
values.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2019-12-16'
__license__ = 'GPLv3+'

import copy

import numpy as np

from viperleed.calc.classes.atom_containers import AtomList
from viperleed.calc.classes.rparams import Rparams
from viperleed.calc.classes.slab import NoBulkRepeatError
from viperleed.calc.files import poscar
from viperleed.calc.lib.woods_notation import writeWoodsNotation
from viperleed.cli_base import ViPErLEEDCLI


_BULK_DIST_ENOUGH = 1.2
_BULK_DIST_SMALL = 0.7


def ask_user_bulk_cut():
    """Return a valid fractional cut position after asking the user."""
    cut = -1.0
    while cut < 0:
        cutstr = input('Enter c value at which bulk starts: ')
        if not cutstr:
            continue
        try:
            cut = float(cutstr)
        except ValueError:
            print('Failed to convert input to float.')
            continue
        if not 0 <= cut < 1:
            print('Error: c value has to be between 0 and 1.')
            cut = -1.0
    return cut



def ask_user_confirmation_to_cut(bulk_interlayer, bulk_cuts):
    """Ask confirmation if cutting at a somewhat narrow layer spacing."""
    if not _BULK_DIST_SMALL < bulk_interlayer < _BULK_DIST_ENOUGH:
        return

    dec = None
    while dec is None:
        reply = input('Cutting the bulk into two layers is possible '
                      f'with a spacing of {round(bulk_interlayer, 2)} A. '
                      'Proceed? (y/[n]): ')
        reply = reply.lower()
        if not reply or reply.startswith('n'):
            dec = False
        elif reply.startswith('y'):
            dec = True
    if not dec:
        bulk_cuts.pop()  # Keep only the top one


def ask_user_symmetry_eps():
    """Return tolerance for symmetry search after asking the user."""
    eps = -1.0
    while eps < 0:
        reply = input('Enter tolerances for symmetry search '
                      '(Default: [0.1 A]): ')
        if not reply:
            eps = 0.1
            break
        try:
            eps = float(reply)
        except ValueError:
            print('Could not convert value to float.')
            continue
        if eps < 0:
            print('Value must be non-negative.')
    return eps


def print_utility_description():
    """Print some introductory information about this utility."""
    print('This utility reads a POSCAR file with arbitrary thickness and '
          'orientation of the bulk, requiring only an input where the bulk '
          'starts. The bulk repeat unit is then automatically determined. '
          'Output is a new POSCAR with a minimum amount of bulk, and values '
          'for the BULK_REPEAT, N_BULK_LAYERS and LAYER_CUTS parameters.\n')


def read_user_poscar():
    """Return a Slab read from a POSCAR file after asking the user."""
    # read the POSCAR file
    filename = ''
    while not filename:
        filename = input('Enter POSCAR file name (Default: [POSCAR]): ')
        if not filename:
            filename = 'POSCAR'
        try:
            slab = poscar.read(filename=filename)
        except FileNotFoundError:
            print(f'File {filename} not found.')
            filename = ''
    print('Slab POSCAR was read successfully.')
    return slab


def write_poscar_min(slab, rpars, cut, bulk_cuts):
    """Write to file a POSCAR with one bulk cell only at the bottom."""
    # Crop off all the atoms at the bottom, leaving only one bulk cell
    slab = copy.deepcopy(slab)
    frac_atoms_z = [at.pos[2] for at in slab]
    frac_topmost_bulk = max(f for f in frac_atoms_z if f <= cut)
    frac_bottom_slab = min(f for f in frac_atoms_z if f > cut)
    dist_to_slab = abs(frac_bottom_slab - frac_topmost_bulk)
    frac_bulk_c = abs(np.linalg.inv(slab.ucell).dot(rpars.BULK_REPEAT)[2])

    new_zero = frac_topmost_bulk + dist_to_slab / 2 - frac_bulk_c
    slab.atlist = AtomList(at for at in slab if at.pos[2] > new_zero)
    slab.update_element_count()   # update the number of atoms per element

    # Shift atoms down, then scale the unit-cell height
    for atom in slab:
        atom.pos[2] -= new_zero
    slab.update_cartesian_from_fractional()
    slab.c_vector[:] *= 1 - new_zero
    slab.update_fractional_from_cartesian()

    # Finally, convert the cut positions
    bulk_cuts = [round(c - new_zero, 3) for c in bulk_cuts]

    # write POSCAR_min
    slab.sort_original()
    try:
        poscar.write(slab, filename='POSCAR_min', comments='none')
    except OSError:
        print('Exception occurred while writing POSCAR_min')
    else:
        print('Wrote POSCAR_min, to be used with parameters below.')

    # print info
    print('\nParameters found:')
    print(f'BULK_REPEAT = xyz{rpars.BULK_REPEAT}')
    print(f'N_BULK_LAYERS = {len(bulk_cuts)}')
    print(f'LAYER_CUTS = {" ".join(bulk_cuts)}   ! etc.')
    input('Program finished, exit with return')


class DetectBulkCLI(ViPErLEEDCLI,
                    cli_name='get_bulk_repeat',
                    help_='automatically detect the bulk portion of a POSCAR'):
    """Detect bulk of a user-given POSCAR from a BULK_LIKE_BELOW."""

    def __call__(self, _=None):
        """Execute the core functionality of the utility.

        Returns
        -------
        exit_code : int
            0: successful completion, 1: errors
        """
        print_utility_description()
        try:
            slab = read_user_poscar()
        except OSError:
            print('Exception while reading POSCAR file')
            return 1

        rpars = Rparams()
        rpars.BULK_LIKE_BELOW = cut = ask_user_bulk_cut()
        rpars.SYMMETRY_EPS = rpars.SYMMETRY_EPS.from_value(
            ask_user_symmetry_eps()
            )

        print('Checking bulk unit cell...')
        try:
            bulk_cuts, bulk_interlayer = slab.detect_bulk(rpars,
                                                          _BULK_DIST_SMALL)
        except NoBulkRepeatError:
            _no_repeat = True
        else:
            _no_repeat = False

        superlattice = rpars.SUPERLATTICE
        ws = writeWoodsNotation(superlattice)                                   # TODO: replace writeWoodsNotation with guilib functions
        info = (
            f'= {ws}' if ws
            else 'M = {} {}, {} {}'.format(*superlattice.astype(int).ravel())
            )
        print(f'Found SUPERLATTICE {info}')

        if _no_repeat:
            print('No repeat vector was found inside the bulk. The bulk may '
                  'already be minimal.')
            return 0

        ask_user_confirmation_to_cut(bulk_interlayer, bulk_cuts)

        # write POSCAR_bulk
        try:
            poscar.write(slab.bulkslab,
                         filename='POSCAR_bulk',
                         comments='none')
        except OSError:
            print('Exception occurred while writing POSCAR_bulk')
        else:
            print('Wrote POSCAR_bulk. Check file to see if periodicity is '
                  'correct.')

        write_poscar_min(slab, rpars, cut, bulk_cuts)
        return 0


if __name__ == '__main__':
    DetectBulkCLI.run_as_script()



"""Module base of viperleed.calc.classes.rparams.special.

Defines SpecialParameter, the base class for all other parameter
classes defined in this package, and a few exceptions.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-10-27'
__license__ = 'GPLv3+'


class SpecialParameterError(Exception):
    """Base exception for all special parameter classes."""


class NotASpecialParameterError(SpecialParameterError):
    """Not a known SpecialParameter, or incorrectly subclassed."""

    def __init__(self, param):
        """Initialize exception."""
        super().__init__(f'{param} is not a special parameter, or '
                         'it is not a subclass of SpecialParameter. '
                         f'known_subclasses={SpecialParameter._subclasses}')


class SpecialParameter:
    """Base class for all the PARAMETERS classes defined in this package."""

    _subclasses = {}  # {PARAM_NAME: cls}

    @classmethod
    def __init_subclass__(cls, *, param=None, **kwargs):
        """Register `cls` as being associated with `param`."""
        super().__init_subclass__(**kwargs)
        if param is not None:
            cls._subclasses[param.upper()] = cls

    @classmethod
    def from_value(cls, value):
        """Return an instance of this class from another `value`.

        This method is provided so that subclasses that need to treat
        `value` can reimplement this method. Examples are classes that
        need to tuple-unpack `value`, or that have other convenience
        methods that needs to be called beforehand.
        The default implementation creates an instance of this class
        with value as its only argument.

        Parameters
        ----------
        value : object
            Value to be used to create an instance of this class.

        Returns
        -------
        instance : SpecialParameter
            The instance created.
        """
        return cls(value)

    @classmethod
    def get_subclass(cls, param):
        """Return the subclass for `param`, or complain."""
        param = param.upper()
        try:
            return next(c for p, c in cls._subclasses.items()
                        if p.startswith(param))
        except StopIteration:
            raise NotASpecialParameterError(param) from None



"""ViPErLEED utility: Merge POSCAR files."""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-03'
__license__ = 'GPLv3+'

from copy import deepcopy

import numpy as np
from scipy.spatial.distance import cdist

from viperleed.calc.files import poscar

from viperleed.cli_base import minimum_length, positive_float
from viperleed.utilities.poscar.base import _PoscarStreamCLI


def merge_slabs(slabs, check_collisions=True, eps=1e-5):
    """Return a single slab containing all the atoms from slabs.

    Parameters
    ----------
    slabs : Sequence of Slab
        The slab objects to be merged.
    check_collisions : bool, optional
        Whether collisions between atoms in the slabs to be merged
        should be reported. Default is True.
    eps : float, optional
        Cartesian tolerance (in angstrom) on the distances between
        pairs of atoms to be considered colliding. Default is 1e-5.

    Returns
    -------
    merged_slab : Slab
        A single slab with all the atoms from `slabs`.

    Raises
    ------
    RuntimeError
        If check_collisions is True, and there are atoms closer
        than `eps` when merging `slabs`.

    Notes
    -----
    The atoms of the slabs are currently NOT CORRECTLY UPDATED:
        Their slab, layer, etc. remain those of `slabs`, and
        not the corresponding properties of `merged_slab`.
    """
    merged_slab = deepcopy(slabs[0])
    # Allow duplicate atom.num till we fix
    # them later with update_atom_numbers()
    merged_slab.atlist.strict = False
    for slab in slabs[1:]:
        if check_collisions:
            # check for collisions
            dists = cdist([at.cartpos for at in merged_slab],
                          [at.cartpos for at in slab])
            if np.any(dists < eps):
                raise RuntimeError('Distance between atoms in different '
                                   f'slabs is smaller than eps={eps}.')
        merged_slab.atlist.extend(slab)                                         # TODO: Potential issue: The slab (and layer, etc.) of each added atom is NOT updated. Probably better to use duplicates if this function is to be used anywhere else.

    merged_slab.update_atom_numbers()
    merged_slab.update_element_count()
    merged_slab.update_cartesian_from_fractional()
    merged_slab.atlist.strict = True
    return merged_slab


_EPS_CELL_DEFAULT = 1e-3
_EPS_COLLISION_DEFAULT = 0.1


class MergePoscarsCLI(_PoscarStreamCLI, cli_name='merge'):
    """Merge two or more POSCAR files into one."""

    long_name = 'merge POSCAR files'

    def add_parser_arguments(self, parser):
        """Add positional list of files and optional arguments."""
        # This version of the overridden add_parser_arguments is
        # a bit tedious: we do not want to have a single stdin as
        # an input stream, but still want to support:
        # (1) the base arguments of ViPErLEEDCLI: we do this via
        #     super, skipping _PoscarStreamCLI in the inheritance
        super(_PoscarStreamCLI, self).add_parser_arguments(parser)
        # (2) the verbose option for the logger.
        self.add_verbose_option(parser)
        parser.add_argument('files', nargs='+', action=minimum_length(2))
        parser.add_argument(
            '--eps-cell',
            help=('maximum element-wise allowed difference in lattice '
                  'parameters (in angstrom) between the unit cells of '
                  'the POSCAR files to be merged. Default is '
                  f'{_EPS_CELL_DEFAULT:g}'),
            type=positive_float,
            default=_EPS_CELL_DEFAULT,
            )
        parser.add_argument(
            '--eps-collision',
            help=('minimum distance (in angstrom) between atoms in '
                  'different slabs before a collision is detected. '
                  f'Default is {_EPS_COLLISION_DEFAULT}'),
            type=positive_float,
            default=_EPS_COLLISION_DEFAULT
            )
        # (3) an output stream, defaulting to stdout.
        self.add_outfile_argument(parser)

    # DISABLE: the positional argument is renamed to slabs on purpose
    # as this makes it clearer that this method expects multiple slabs,
    # in contrast with the one of the base class. This also will raise
    # if anyone should pass it along as a named argument.
    # pylint: disable-next=arguments-renamed
    def process_slab(self, slabs, args):
        """Merge slabs into a new one."""
        return merge_slabs(slabs,
                           check_collisions=True,
                           eps=args.eps_collision)

    def read_poscar(self, args):
        """Read POSCAR files, checking that they are consistent."""
        slabs = [poscar.read(file) for file in args.files]

        # Check that all pairs of unit cells match within the tolerance
        for i, slab in enumerate(slabs[:-1]):
            this_ucell = slab.ucell
            other_ucells = [other_slab.ucell for other_slab in slabs[i:]]
            if not np.allclose(other_ucells, this_ucell, atol=args.eps_cell):
                self.parser.error('Files do not have the same unit cell. '
                                  'Use --eps-cell to increase tolerance.')
        return slabs


if __name__ == '__main__':
    MergePoscarsCLI.run_as_script()



"""Module run_sections of viperleed.calc.sections.

Wrapper functions for running a single section as well as
a section loop.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'

import logging
import os
import time
from timeit import default_timer as timer

from viperleed.calc.files import beams as iobeams
from viperleed.calc.files import parameters
from viperleed.calc.files.displacements import readDISPLACEMENTS
from viperleed.calc.files.phaseshifts import readPHASESHIFTS
from viperleed.calc.files.vibrocc import readVIBROCC, writeVIBROCC
from viperleed.calc.lib.base import get_elapsed_time_str
from viperleed.calc.sections import deltas
from viperleed.calc.sections import errorcalc
from viperleed.calc.sections import fd_optimization
from viperleed.calc.sections import initialization                              # TODO: perhaps use CalcSection
from viperleed.calc.sections import refcalc
from viperleed.calc.sections import rfactor
from viperleed.calc.sections import search
from viperleed.calc.sections import superpos
from viperleed.calc.sections.cleanup import cleanup, move_oldruns


logger = logging.getLogger(__name__)


def run_section(index, sl, rp):
    """Run a specific viperleed.calc section.

    Parameters
    ----------
    index : int
        Index of the section, see sectionNames variable below.
    sl : Slab
        Slab object containing atom information.
    rp : Rparams
        The run parameters.

    Returns
    -------
    None.
    """
    sectionNames = {0: "INITIALIZATION",
                    1: "REFERENCE CALCULATION",
                    2: "DELTA-AMPLITUDES",
                    3: "SEARCH",
                    11: "R-FACTOR CALCULATION",
                    12: "R-FACTOR CALCULATION",
                    31: "SUPERPOS",
                    5:  "ERROR CALCULATION",
                    6:  "FULL-DYNAMIC OPTIMIZATION"}
    # files that need to be there for the different parts to run
    requiredFiles = {0: ["POSCAR", "PARAMETERS", "VIBROCC", "IVBEAMS"],
                     1: ["BEAMLIST", "PHASESHIFTS", "POSCAR", "PARAMETERS",
                         "IVBEAMS", "VIBROCC"],
                     2: ["BEAMLIST", "PHASESHIFTS", "POSCAR", "PARAMETERS",
                         "IVBEAMS", "VIBROCC", "DISPLACEMENTS"],
                     3: ["BEAMLIST", "PHASESHIFTS", "POSCAR", "PARAMETERS",
                         "IVBEAMS", "VIBROCC", "DISPLACEMENTS", "EXPBEAMS"],
                     11: ["BEAMLIST", "PARAMETERS", "IVBEAMS", "EXPBEAMS"],
                     12: ["BEAMLIST", "PARAMETERS", "IVBEAMS", "EXPBEAMS"],
                     31: ["BEAMLIST", "POSCAR", "PARAMETERS", "IVBEAMS",
                          "VIBROCC", "DISPLACEMENTS"],
                     5: ["BEAMLIST", "PHASESHIFTS", "POSCAR", "PARAMETERS",
                         "IVBEAMS", "VIBROCC", "DISPLACEMENTS", "EXPBEAMS"],
                     6: ["BEAMLIST", "PHASESHIFTS", "POSCAR", "PARAMETERS",
                         "IVBEAMS", "VIBROCC", "EXPBEAMS"]}

    checkfiles = requiredFiles[index][:]
    o = "\nSTARTING SECTION: "+sectionNames[index]
    if index == 3 and rp.disp_blocks and rp.disp_blocks[rp.search_index][1]:
        o += " "+rp.disp_blocks[rp.search_index][1]  # displacement block name
    if rp.domainParams or rp.DOMAINS:
        o += " (DOMAINS)"
        for fn in ["POSCAR", "VIBROCC", "PHASESHIFTS"]:
            try:
                checkfiles.remove(fn)
            except Exception:
                pass
    logger.info(o)
    sectionStartTime = timer()
    rp.runHistory.append(index)
    for dp in rp.domainParams:
        dp.rp.runHistory = rp.runHistory
    i = 0
    while i < len(checkfiles):
        filename = checkfiles[i]
        ignoreError = False
        if rp.fileLoaded[filename]:
            i += 1
            continue
        # try loading files
        if filename == "EXPBEAMS":
            rp.try_loading_expbeams_file()
            if index != 0:
                iobeams.checkEXPBEAMS(sl, rp)
        elif filename == "IVBEAMS":
            try:
                rp.ivbeams = iobeams.readIVBEAMS()
                rp.ivbeams_sorted = False
                rp.fileLoaded["IVBEAMS"] = True
            except FileNotFoundError:
                if (os.path.isfile("EXPBEAMS")
                        or os.path.isfile("EXPBEAMS.csv")):
                    checkfiles.insert(i+1, "EXPBEAMS")
                    logger.warning("IVBEAMS file not found. Will attempt "
                                   "generating IVBEAMS from EXBEAMS.")
                    ignoreError = True
                else:
                    logger.error("Neither IVBEAMS nor EXPBEAMS file "
                                 "found.")
            except Exception as e:
                logger.error("Error while reading required file IVBEAMS",
                             exc_info=(type(e) != FileNotFoundError))
        elif filename == "BEAMLIST":
            try:
                rp.beamlist = iobeams.readBEAMLIST()
                rp.fileLoaded["BEAMLIST"] = True
            except Exception as e:
                logger.error("Error while reading required file "
                             "BEAMLIST", exc_info=(type(e) !=
                                                   FileNotFoundError))
        elif filename == "VIBROCC":
            changeVIBROCC = False
            try:
                changeVIBROCC = readVIBROCC(rp, sl)
                rp.fileLoaded["VIBROCC"] = True
            except Exception as exc:
                _not_found = not isinstance(exc, FileNotFoundError)
                logger.error("Error while reading required file VIBROCC",
                             exc_info=_not_found)
            sl.full_update(rp)
            if changeVIBROCC:
                if os.path.isfile("VIBROCC"):
                    os.rename("VIBROCC", "VIBROCC_user")
                    rp.manifest.append("VIBROCC_user")
                    logger.info(
                        "VIBROCC file was modified with automatically "
                        "generated vibrational amplitudes."
                        )
                writeVIBROCC(sl, rp, "VIBROCC")
                rp.manifest.append("VIBROCC")
            if rp.T_EXPERIMENT is not None:
                parameters.comment_out(rp, "T_EXPERIMENT")
            if rp.T_DEBYE is not None:
                parameters.comment_out(rp, "T_DEBYE")
            if len(rp.VIBR_AMP_SCALE) > 0:
                parameters.comment_out(rp, "VIBR_AMP_SCALE")
        elif filename == "PHASESHIFTS":
            try:
                (rp.phaseshifts_firstline, rp.phaseshifts,
                 newpsGen, newpsWrite) = readPHASESHIFTS(sl, rp)
                if newpsGen:
                    logger.error(
                        "PHASESHIFTS file generation is only supported "
                        "during initialization. Stopping execution...")
                    raise RuntimeError("Inconsistent PHASESHIFT file")
                elif newpsWrite:
                    logger.warning(
                        "Writing a new PHASESHIFTS file is "
                        "only supported during initialization. The "
                        "data in the provided file will be used, but "
                        "running the initialization is recommended.")
                    rp.fileLoaded["PHASESHIFTS"] = True
                else:
                    rp.fileLoaded["PHASESHIFTS"] = True
            except Exception as e:
                logger.error("Error while reading required file "
                             "PHASESHIFTS", exc_info=(type(e) !=
                                                      FileNotFoundError))
        elif filename == "DISPLACEMENTS":
            try:
                readDISPLACEMENTS(rp)
                rp.fileLoaded["DISPLACEMENTS"] = True
            except Exception as e:
                logger.error("Error while reading required file "
                             "DISPLACEMENTS", exc_info=(type(e) !=
                                                        FileNotFoundError))
        if not rp.fileLoaded[filename] and not ignoreError:
            # and if that didn't work, stop:
            logger.error(f"Step '{sectionNames[index]}' requires "
                         f"file {filename}. Stopping execution...")
            rp.setHaltingLevel(3)
            return
        i += 1
    try:
        if index == 0:
            initialization.initialization(sl, rp)
        elif index == 1:
            refcalc.refcalc(sl, rp)
        elif index in [11, 12]:
            rfactor.rfactor(sl, rp, index)
        elif index == 2:
            deltas.deltas(sl, rp)
        elif index == 3:
            search.search(sl, rp)
        elif index == 31:
            superpos.superpos(sl, rp)
        elif index == 5:
            errorcalc.errorcalc(sl, rp)
        elif index == 6:
            fd_optimization.fd_optimization(sl, rp)
    except Exception:
        logger.error(f"Error in section {sectionNames[index]}")
        raise
    logger.info(
        f"Finishing section at {time.strftime('%H:%M:%S', time.localtime())}. "
        f"Section took {get_elapsed_time_str(timer() - sectionStartTime)}."
        )


def section_loop(rp, sl):
    """
    Executes sections as specified in rp.RUN, may loop if required.

    Parameters
    ----------
    rp : Rparams
        The run parameters.
    sl : Slab
        Slab object containing atom information.

    Returns
    -------
    int
        0: exit without errors.
        1: clean exit through KeyboardInterrupt
        2: exit due to Exception before entering main loop
        3: exit due to Exception during main loop
    """
    sectionorder = [0, 1, 6, 11, 2, 3, 31, 12, 4, 5]
    searchLoopR = None
    searchLoopLevel = 0
    initHalt = False
    while rp.RUN:
        try:
            sec = rp.RUN.pop(0)
            if rp.runHistory and (sectionorder.index(sec)
                                  < sectionorder.index(rp.runHistory[-1])):
                logger.info("\nExecution repeats. Moving old output to "
                            "workhistory folder.")
                try:
                    move_oldruns(rp)
                except Exception:
                    logger.warning(
                        "Exception while trying to clean up earlier segments. "
                        "Program will proceed, but old files may be lost.",
                        exc_info=True
                        )
            run_section(sec, sl, rp)
            if rp.domainParams and sl is None:                                  # is there any point in allowing sl to be None?
                sl = rp.pseudoSlab
            if rp.domainParams:
                rp.setHaltingLevel(max(dp.rp.halt for dp in rp.domainParams))

            # Decide how to proceed
            next_section = next(iter(rp.RUN), None)
            if (sec == 0 and not rp.domainParams and not sl.preprocessed
                    and rp.HALTING <= 2 and len(rp.RUN) > 0):
                logger.info("Initialization finished. Execution will stop. "
                            "Please check whether comments in POSCAR are "
                            "correct, then restart.")
                rp.checklist.append("Check whether comments in "
                                    "POSCAR are correct")
                rp.setHaltingLevel(2)
                initHalt = True
            elif (sec == 1 and rp.fileLoaded["EXPBEAMS"]):
                if (next_section != 11 and          # r-factor after refcalc
                        (not rp.domainParams or 3 in rp.runHistory)):
                    rp.RUN.insert(0, 11)
            elif (sec == 3 and rp.fileLoaded["EXPBEAMS"]):
                if next_section != 31:  # superpos after search
                    rp.RUN.insert(0, 31)
            elif sec == 31 and rp.fileLoaded["EXPBEAMS"]:
                if next_section != 12:   # r-factor after superpos
                    rp.RUN.insert(0, 12)
            elif sec == 12 and not rp.STOP:
                loops = [t for t in rp.disp_loops if t[1] == rp.search_index]
                if loops:
                    if searchLoopLevel == 0 or searchLoopR > rp.last_R:
                        searchLoopR = rp.last_R
                        searchLoopLevel = len(loops)
                    elif searchLoopR <= rp.last_R:
                        searchLoopLevel -= 1
                    if searchLoopLevel != 0:
                        rp.search_index = sorted(loops)[searchLoopLevel-1][0]
                        logger.info("Search loop: repeating at block "
                                    + rp.disp_blocks[rp.search_index][1])
                    else:
                        rp.search_index += 1
                        o = "Search loop ends."
                        if len(rp.disp_blocks) > rp.search_index:
                            o += (" Continuing at block "
                                  + rp.disp_blocks[rp.search_index][1])
                        logger.info(o)
                else:
                    rp.search_index += 1
                for dp in rp.domainParams:
                    dp.rp.search_index = rp.search_index
                if len(rp.disp_blocks) > rp.search_index:
                    if not rp.domainParams:
                        sl.restoreOriState()
                    rp.resetSearchConv()
                    for dp in rp.domainParams:
                        dp.sl.restoreOriState()
                        dp.rp.resetSearchConv()
                    if rp.RUN[:2] != [2, 3]:
                        rp.RUN = [2, 3] + rp.RUN
        except KeyboardInterrupt:
            logger.warning("Stopped by keyboard interrupt, attempting "
                           "clean exit...")
            cleanup(rp.manifest, rp)
            return 1
        except Exception:
            logger.error("Exception during viperleed.calc execution: ",
                         exc_info=True)
            cleanup(rp.manifest, rp)
            return 3
        if rp.halt >= rp.HALTING:
            if not initHalt:
                logger.info(
                    "# An exception occurred that meets the halting "
                    "criteria defined by the HALTING parameter. Execution "
                    "will stop, check log for warnings and errors."
                    )
            break
        parameters.update(rp)  # Look for a user STOP
        if rp.RUN and rp.STOP and not rp.RUN[0] in [11, 12, 31]:
            logger.info("# Stopped by user STOP command.")
            break
    cleanup(rp.manifest, rp)
    return 0



"""Module delta_intensities of viperleed.calc.files.

Reads in delta-amplitude files.
"""

__authors__ = (
    'Tobias Hable (@ElHablos)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2022-05-02'
__license__ = 'GPLv3+'

import os
from re import I
import sys

import fortranformat as ff
import matplotlib.pyplot as plt
from numba import njit, prange
import numpy as np
from numpy import cos, sin, sqrt
import scipy
from tqdm import tqdm


def read_delta_file(filename, n_E):
    """This function reads in one file of data and stores the data in arrays, which can be used in later functions (ex.: GetInt)

    Parameters
    ----------
    filename : string
    The filename describes which file you want to read in (path from the function location to the file)

    n_E : int
    Number of different energy levels. This should be a known factor for your files


    Returns
    -------
    (phi, theta): tuple of float
    Angles of how the beam hits the sample

    (trar1, trar2) : tuple of ndarray
    Vectors of the normal and the reciprocal unit cell of the sample

    int0 : int
    Number of beams that are reflected by the sample

    n_atoms: int
    TBD

    nc_steps: ndarray
    Number of permutations between direction deltas and vibration deltas

    E_kin_array : ndarray
    Array that contains the kinetic energies of the elastically scatted
    electrons inside the crystal. (Not the incidence energy!)

    VPI_array : ndarray
    Imaginary part of the inner potential of the surface

    VV_array : ndarray
    Real part of the inner potential of the surface

    beam_indices : ndarray
    Array with the order of beams

    Cundisp : ndarray
    TBD always 0

    CDisp : ndarry
    Geometric displacement of given delta

    amplitudes_ref : ndarray
    Array that contains all values of the reference amplitudes

    amplitudes_del : ndarray
    Array that contains all values of the delta amplitudes
    """

    # Lists and Arrays needed; E,VPI,VV are numpy arrays that get returned, the others are lists that help saving the other data
    HeaderBlock1 = []
    HeaderBlock2 = []
    E_kin_array = np.full(n_E, fill_value=np.NaN)
    VPI_array = np.full(n_E, fill_value=np.NaN)
    VV_array = np.full(n_E, fill_value=np.NaN)
    listdummy = []
    listdummy2 = []

    # we need two fortran format readers
    ff_reader_6E13_7 = ff.FortranRecordReader("6E13.7")
    ff_reader_10F7_4 = ff.FortranRecordReader("10F7.4")

    # Reading in the data of a file
    with open(filename, mode="r") as file:
        content = file.readlines()
    # make into an iterator
    file_lines = iter(content)

    # 1.Block of Header - only 1 line - theta, phi, trar1, trar2 variables
    line = next(file_lines)
    HeaderBlock1 = ff_reader_6E13_7.read(line)
    trar1 = [0, 0]
    trar2 = [0, 0]
    theta, phi, trar1[0], trar1[1], trar2[0], trar2[1] = HeaderBlock1
    trar1 = np.array(trar1)
    trar2 = np.array(trar2)

    # 2.Block of Header - also only 1 line - int0, n_atoms, nc_steps variables
    line = next(file_lines)
    z2_elemente = line.split()
    for part in z2_elemente:
        HeaderBlock2.append(int(part))
    int0, n_atoms, nc_steps = HeaderBlock2

    # 3.Block of Header - Positions of the beams
    while len(listdummy2) < 2 * int0:
        line = next(file_lines)
        listdummy = line.split()
        for part in listdummy:
            listdummy2.append(part)
        # dieses if wieder redundant?
        if len(listdummy2) >= 2 * int0:
            beam_indices = np.full(shape=[int0, 2], fill_value=np.NaN)
            for i in range(int0):
                for j in range(2):
                    beam_indices[i, j] = listdummy2[2 * i + j]
    listdummy.clear()
    listdummy2.clear()

    # 4.Block of Header - Cundisp
    while len(listdummy2) < 3 * n_atoms:
        line = next(file_lines)
        listdummy = line.split()
        for part in listdummy:
            listdummy2.append(part)
        # if redundant?
        if len(listdummy2) >= 3 * n_atoms:
            Cundisp = np.full(shape=[n_atoms, 3], fill_value=np.NaN)
            for i in range(n_atoms):
                for j in range(3):
                    Cundisp[i, j] = listdummy2[3 * i + j]
    listdummy.clear()
    listdummy2.clear()

    # 5.Block of Header - CDisp
    while len(listdummy2) < 3 * n_atoms * nc_steps:
        line = next(file_lines)
        listdummy = ff_reader_10F7_4.read(line)
        for part in listdummy:
            listdummy2.append(part)
            # maybe getting rid of the if again
        if len(listdummy2) >= 3 * n_atoms * nc_steps:
            CDisp = np.full(shape=[nc_steps, n_atoms, 3], fill_value=np.NaN)
            for j in range(nc_steps):  # some syntax error here
                for k in range(n_atoms):
                    for l in range(3):
                        CDisp[j, k, l] = listdummy2[n_atoms * 3 * j + 3 * k + l]
    listdummy.clear()
    listdummy2.clear()

    # 6.Block of Header - Aid
    while len(listdummy2) < nc_steps:
        line = next(file_lines)
        listdummy = ff_reader_10F7_4.read(line)
        for part in listdummy:
            listdummy2.append(part)
        if len(listdummy2) >= nc_steps:
            Aid = np.full(shape=[nc_steps], fill_value=np.NaN)
            for i in range(nc_steps):
                Aid[i] = listdummy2[i]
    listdummy.clear()
    listdummy2.clear()

    # Initialize arrays for reference and delta amplitudes
    amplitudes_ref = np.full(shape=[n_E, int0], fill_value=np.nan, dtype=complex)
    amplitudes_del = np.full(
        shape=[n_E, nc_steps, int0], fill_value=np.nan, dtype=complex
    )

    # maybe working arrays for transfer into amplitude arrays ?

    # End of the Header - Start of Reading in the Delta Data
    for e_index in range(n_E):  # Energy loop

        # Energy, VPI and VV header
        line = next(file_lines)
        listdummy = ff_reader_6E13_7.read(line)
        for part in listdummy:
            if part is not None:
                listdummy2.append(part)
        E_kin, VPI, VV = listdummy2
        # transversing  Hartree to eV
        E_kin = hartree_to_eV(E_kin)
        E_kin_array[e_index] = E_kin
        VPI_array[e_index] = VPI
        VV_array[e_index] = VV
        listdummy.clear()
        listdummy2.clear()

        # Reference amplitudes
        while len(listdummy2) < 2 * int0:
            line = next(file_lines)
            listdummy = ff_reader_6E13_7.read(line)
            for part in listdummy:
                listdummy2.append(part)
        for j in range(int0):
            amplitudes_ref[e_index, j] = complex(
                listdummy2[2 * j], listdummy2[2 * j + 1]
            )
        listdummy.clear()
        listdummy2.clear()

        # Delta amplitudes
        while len(listdummy2) < 2 * int0 * nc_steps:
            line = next(file_lines)
            listdummy = ff_reader_6E13_7.read(line)
            for part in listdummy:
                listdummy2.append(part)
        for j in range(nc_steps):
            for k in range(int0):
                amplitudes_del[e_index, j, k] = complex(
                    listdummy2[2 * int0 * j + k * 2],
                    listdummy2[2 * int0 * j + k * 2 + 1],
                )
        listdummy.clear()
        listdummy2.clear()

    return (
        (phi, theta),
        (trar1, trar2),
        (int0, n_atoms, nc_steps),
        beam_indices,
        Cundisp,
        CDisp,
        (E_kin_array, VPI_array, VV_array),
        amplitudes_ref,
        amplitudes_del,
    )


@njit(fastmath=True, parallel=True, nogil=True)
def calc_delta_intensities(
    phi,
    theta,
    trar1,
    trar2,
    Beam_variables,
    beam_indices,
    ph_CDisp,
    E_array,
    VPI_array,
    VV_array,
    amplitudes_ref,
    amplitudes_del,
    n_files,
    nc_surf,
    delta_steps,
    number_z_steps,
):
    """This function reads in the values of the function Transform and uses them to get the ATSAS_matrix

    Parameters
    ----------
    n_files: int
    Number of files

    phi, theta:
    Angles of how the beam hits the sample

    trar1, trar2:
    Vectors of the normal and the reciprocal unit cell of the sample

    Beam_variables:
    The variables int0, n_atoms, nc_steps for each file stored in an array

    beam_indices:
    Array with the order of beams

    ph_CDisp:
    Geometric displacements of the atom

    E_array:
    Array that contains all the energies of the file

    VPI_array:
    Imaginary part of the inner potential of the surface

    VV_array:
    Real part of the inner potential of the surface

    amplitudes_ref:
    Array that contains all values of the reference amplitudes

    amplitudes_del:
    Array that contains all values of the delta amplitudes

    nc_surf: np.array of bool
    Bool array with flags that decide if atom is considered to be at the surface.

    delta_steps:
    List of numbers that decide which geometric displacement this atom has

    number_z_steps:
    Total number of different delta_z values


    Returns
    ----------
    ATSAS_matrix:
    Array that contains the intensities of the different beams with each energy
    """

    # Conc will probably taken as Input parameter aswell
    CXDisp = 1000
    XDisp = 0
    Conc = 1
    for i in range(n_files):
        if nc_surf[i]:
            int0, n_atoms, nc_steps = Beam_variables[i, :]
            for j in range(n_atoms):
                delta_fraction = delta_steps[i]  # probably besserer Name als fill finden
                delta_step_left = int(delta_fraction // 1)
                # necessary so that the x2 doesnt go out of the index range
                if delta_step_left == number_z_steps - 1:
                    delta_step_left = delta_step_left - 1
                delta_step_right = delta_step_left + 1
                x = delta_fraction - delta_step_left
                del_left = ph_CDisp[i, delta_step_left, j, 0]
                del_right = ph_CDisp[i, delta_step_right, j, 0]
                CDisp = x * (del_right - del_left) + del_left
                XDisp = Conc * CDisp
                if XDisp < CXDisp:
                    CXDisp = XDisp

    ATSAS_matrix = np.zeros((len(E_array), int0))

    # Loop über die Energien
    for e_index in prange(len(E_array)):
        # Definieren von Variablen, die in der jeweiligen Energie gleichbleiben
        E = E_array[e_index]
        VV = VV_array[e_index]
        VPI = VPI_array[e_index]

        AK = sqrt(max(2 * E - 2 * VV, 0))
        C = AK * cos(theta)
        BK2 = AK * sin(theta) * cos(phi)
        BK3 = AK * sin(theta) * sin(phi)
        BKZ = sqrt(complex(2 * E - BK2 ** 2 - BK3 ** 2, -2 * VPI))

        # Loop über die Beams
        for b_index in range(int0):
            # Variablen per Beam
            h = beam_indices[b_index, 0]
            k = beam_indices[b_index, 1]
            AK2 = BK2 + h * trar1[0] + k * trar2[0]
            AK3 = BK3 + h * trar1[1] + k * trar2[1]
            AK = 2 * E - AK2 ** 2 - AK3 ** 2
            AKZ = complex(AK, -2 * VPI)
            APERP = AK - 2 * VV

            # Herausfinden welche NCStep deltas man nimmt mit delta_step matrix
            DelAct = amplitudes_ref[e_index, b_index]
            for i in range(n_files):
                # noch genau schauen wie genau gewollt
                # Interpolation of float delta_step values
                delta_fraction = delta_steps[i]
                delta_step_left = int(delta_fraction)
                delta_step_right = min(delta_step_left+1, number_z_steps)
                x = delta_fraction - delta_step_left
                del_left = amplitudes_del[i, e_index, delta_step_left, b_index]
                del_right = amplitudes_del[i, e_index, delta_step_right, b_index]
                # 1D interpolation formula
                intpol_value = x * (del_right - del_left) + del_left
                # ###
                DelAct += intpol_value

            if APERP > 0:
                A = sqrt(APERP)
                PRE = (BKZ + AKZ) * CXDisp
                PRE = np.exp(complex(0, -1) * PRE)
                amp_abs = abs(DelAct)
                if amp_abs > 10e10:
                    ATSAS = 10e20
                else:
                    ATSAS = amp_abs ** 2 * abs(PRE) ** 2 * A / C
            else:
                ATSAS = 0

            ATSAS_matrix[e_index, b_index] = ATSAS

    return ATSAS_matrix

@njit(fastmath = True, nogil = True)
def bilinear_interpolation_unit_square(x_vals, y_vals, func_vals):
    x1, x2, x = x_vals
    y1, y2, y = y_vals
    f11, f12, f21, f22 = func_vals

    return (x2-x)*(f11*(y2-y) + f12*(y-y1)) + (x-x1)*(f21*(y2-y) + f22*(y-y1))


@njit(fastmath=True, parallel=True, nogil=True)
def calc_delta_intensities_2D(
    phi,
    theta,
    trar1,
    trar2,
    Beam_variables,
    beam_indices,
    ph_CDisp,
    E_array,
    VPI_array,
    VV_array,
    amplitudes_ref,
    amplitudes_del,
    n_files,
    nc_surf,
    delta_steps,
    number_z_steps,
    number_vib_steps,
):

    """This function reads in the values of the function Transform and uses them to get the ATSAS_matrix

    Parameters
    ----------
    phi, theta:
    Angles of how the beam hits the sample

    trar1, trar2:
    Vectors of the normal and the reciprocal unit cell of the sample

    Beam_variables:
    The variables int0, n_atoms, nc_steps for each file stored in an array

    beam_indices:
    Array with the order of beams

    ph_CDisp:
    Geometric displacements of the atom

    E_array:
    Array that contains all the energies of the file

    VPI_array:
    Imaginary part of the inner potential of the surface

    VV_array:
    Real part of the inner potential of the surface

    amplitudes_ref:
    Array that contains all values of the reference amplitudes

    amplitudes_del:
    Array that contains all values of the delta amplitudes

    n_files: int
    Number of files

    nc_surf: np.array of bool
    Bool array with flags that decide if atom is considered to be at the surface.

    delta_step:
    List of numbers that decide which geometric displacement this atom has

    number_z_steps:
    Total number of different delta_z values

    number_vib_steps:
    Total number of different delta_vib values

    Returns
    ----------
    ATSAS_matrix:
    Array that contains the intensities of the different beams with each energy
    """

    # Conc will probably taken in as an Input parameter
    CXDisp = 1000
    XDisp = 0
    Conc = 1
    for i in range(n_files):
        if nc_surf[i]:
            int0, n_atoms, nc_steps = Beam_variables[i, :]
            for j in range(n_atoms):
                z_step = delta_steps[
                    i, 0
                ]
                x1 = int(z_step)
                x2 = min(x1+1, number_z_steps)
                x = z_step - x1
                y1 = ph_CDisp[i, x1, j, 0]
                y2 = ph_CDisp[i, x2, j, 0]
                CDisp = x * (y2 - y1) + y1
                XDisp = Conc * CDisp
                if XDisp < CXDisp:
                    CXDisp = XDisp

    ATSAS_matrix = np.zeros((len(E_array), int0))

    # Loop over the energies
    for e_index in prange(len(E_array)):
        # defining variables that stay the same for their respective energy
        E = E_array[e_index]
        VV = VV_array[e_index]
        VPI = VPI_array[e_index]

        AK = sqrt(max(2 * E - 2 * VV, 0))
        C = AK * cos(theta)
        BK2 = AK * sin(theta) * cos(phi)
        BK3 = AK * sin(theta) * sin(phi)
        BKZ = sqrt(complex(2 * E - BK2 ** 2 - BK3 ** 2, -2 * VPI))

        # Loop over the beams
        for b_index in range(int0):
            # variables that differ with each beam
            h = beam_indices[b_index, 0]
            k = beam_indices[b_index, 1]
            AK2 = BK2 + h * trar1[0] + k * trar2[0]
            AK3 = BK3 + h * trar1[1] + k * trar2[1]
            AK = 2 * E - AK2 ** 2 - AK3 ** 2
            AKZ = complex(AK, -2 * VPI)
            APERP = AK - 2 * VV

            # finding out which NCStep you take with the delta_step matrix
            DelAct = amplitudes_ref[e_index, b_index]
            for i in range(n_files):
                # Interpolation of the float delta_step values

                z = delta_steps[i, 0]
                v = delta_steps[i, 1]
                z1 = int(z)
                z2 = min(z1+1, number_z_steps)
                v1 = int(v)
                v2 = min(v1+1, number_vib_steps)

                da_z1_v1 = amplitudes_del[i, e_index, v1 * number_vib_steps + z1, b_index]
                da_z1_v2 = amplitudes_del[i, e_index, v2 * number_vib_steps + z1, b_index]
                da_z2_v1 = amplitudes_del[i, e_index, v1 * number_vib_steps + z2, b_index]
                da_z2_v2 = amplitudes_del[i, e_index, v2 * number_vib_steps + z2, b_index]

                intpol_value = bilinear_interpolation_unit_square(
                    (z1, z2, z),
                    (v1, v2, v),
                    (da_z1_v1, da_z1_v2, da_z2_v1, da_z2_v2)
                )
                DelAct += intpol_value

            if APERP > 0:
                A = sqrt(APERP)
                PRE = (BKZ + AKZ) * CXDisp
                PRE = np.exp(complex(0, -1) * PRE)
                amp_abs = abs(DelAct)
                if amp_abs > 10e10:
                    ATSAS = 10e20
                else:
                    RPRE = abs(PRE)
                    ATSAS = amp_abs ** 2 * RPRE ** 2 * A / C
            else:
                ATSAS = 0

            ATSAS_matrix[e_index, b_index] = ATSAS

    return ATSAS_matrix


def hartree_to_eV(E_hartree):
    return E_hartree * 27.211396


def Transform(n_E, directory):
    """This function transforms the read in data to a form, where it can be read in by the GetInt function

    Parameters
    ----------
    n_E:
    Number of different energies of one file

    directory:
    Relative path to the file that gets read in


    Returns
    ----------
    phi, theta:
    Angles of how the beam hits the sample

    trar1, trar2:
    Vectors of the normal and the reciprocal unit cell of the sample

    Beam_variables:
    The variables int0, n_atoms, nc_steps for each file stored in an array

    beam_indices:
    Array with the order of beams

    CDisp:
    Geometric displacements of the atom

    E_array:
    Array that contains all the energies of the file

    VPI_array:
    Imaginary part of the inner potential of the surface

    VV_array:
    Real part of the inner potential of the surface

    amplitudes_ref:
    Array that contains all values of the reference amplitudes

    amplitudes_del:
    Array that contains all values of the delta amplitudes

    filename_list:
    List of the filenames that contain the data
    """

    filename_list = []
    data_list_all = {}
    for filenames in os.walk(directory):
        filename_list.append(filenames)
    filename_list = filename_list[0][2]

    for name in tqdm(filename_list):
        filename = directory + name
        data_list_all[name] = read_delta_file(filename, n_E)

    # constant variables
    phi, theta = data_list_all[filename_list[0]][0]
    trar1, trar2 = data_list_all[filename_list[0]][1]
    beam_indices = data_list_all[filename_list[0]][3]
    E_array, VPI_array, VV_array = data_list_all[filename_list[0]][6]
    amplitudes_ref = data_list_all[filename_list[0]][7]
    n_files = len(filename_list)

    # int0, n_atoms, nc_steps in an array(nc_steps can change, n_atoms maybe too)
    Beam_variables = np.full(shape=[n_files, 3], fill_value=np.NaN, dtype=int)
    for i, name in enumerate(filename_list):
        Beam_variables[i, :] = data_list_all[name][2]

    int0 = int(np.max(Beam_variables[:, 0]))
    n_atoms_max = int(np.max(Beam_variables[:, 1]))
    nc_steps_max = int(np.max(Beam_variables[:, 2]))

    # saving the changing data in arrays
    CDisp = np.full(shape=(n_files, nc_steps_max, n_atoms_max, 3), fill_value=np.NaN)
    amplitudes_del = np.full(
        shape=(n_files, n_E, nc_steps_max, int0), fill_value=np.NaN, dtype=complex
    )
    for i, name in enumerate(filename_list):
        int0, n_atoms, nc_steps = Beam_variables[i]
        int0 = int(int0)
        n_atoms = int(n_atoms)
        nc_steps = int(nc_steps)
        CDisp[i, 0:nc_steps, 0:n_atoms, :] = data_list_all[name][5]
        amplitudes_del[i, :, 0:nc_steps, :] = data_list_all[name][8]

    return (
        phi,
        theta,
        trar1,
        trar2,
        Beam_variables,
        beam_indices,
        CDisp,
        E_array,
        VPI_array,
        VV_array,
        amplitudes_ref,
        amplitudes_del,
        filename_list,
    )


def PlotMaker(
    phi,
    theta,
    trar1,
    trar2,
    Beam_variables,
    beam_indices,
    CDisp,
    E_array,
    VPI_array,
    VV_array,
    amplitudes_ref,
    amplitudes_del,
    n_files,
    NCSurf,
    delta_step_einzeln,
    phi_g,
    theta_g,
    trar1_g,
    trar2_g,
    Beam_variables_g,
    beam_indices_g,
    CDisp_g,
    E_array_g,
    VPI_array_g,
    VV_array_g,
    amplitudes_ref_g,
    amplitudes_del_g,
    n_files_g,
    delta_step_g,
    int0,
    number_z_steps,
    number_vib_steps,
):

    # This function determines the difference between a file, where 2 contributors to the intensities are considered, and 2 files, where only one of these
    # contributors each is considered. That way we can determine, if it is viable to use linear combination of the 3 dimensions + the vib files to describe
    # each delta of the atom.
    #
    # INPUT:
    #
    # phi, theta, phi_g, theta_g:
    # Angles of how the beam hits the sample
    #
    # trar1, trar2, trar1_g, trar2_g:
    # Vectors of the normal and the reciprocal unit cell of the sample
    #
    # Beam_variables, Beam_variables_g:
    # The variables int0, n_atoms, nc_steps for each file stored in an array
    #
    # beam_indices, beam_indices_g:
    # Array with the order of beams
    #
    # CDisp, CDisp_g:
    # Geometric displacements of the atom
    #
    # E_array, E_array_g:
    # Array that contains all the energies of the file
    #
    # VPI_array, VPI_array_g:
    # Imaginary part of the inner potential of the surface
    #
    # VV_array, VV_array_g:
    # Real part of the inner potential of the surface
    #
    # amplitudes_ref, amplitudes_ref_g:
    # Array that contains all values of the reference amplitudes
    #
    # amplitudes_del, amplitudes_del_g:
    # Array that contains all values of the delta amplitudes
    #
    # filename_list, filename_list_g:
    # List of the filenames that contain the data
    #
    # NCSurf:
    # List of 0 and 1 to decide which file takes part in creating the CXDisp
    #
    # delta_step, delta_step_g:
    # List of numbers that decide which geometric displacement this atom has
    #
    # int0:
    # number of beams
    #
    # number_z_steps:
    # Total number of different delta_z values
    #
    # number_vib_steps:
    # Total number of different delta_vib values
    #
    # OUTPUT:
    # ATSAS_matrix:
    # Array that contains the intensities of the different beams with each energy

    matrix_differenz = np.full(shape=(len(E_array), int0), fill_value=np.NaN)
    array = np.full(shape=(101, 101), fill_value=np.NaN)
    for i in tqdm(
        range(101)
    ):  # gerade weird dringend drüber nachdenken, 0.1 schritte ist dann diese range sinnvoll?
        for j in range(101):
            delta_step_einzeln = np.array(
                [j / 10, 5.0, 5.0, 5.0, 5.0, i / 10, 5.0, 5.0, 5.0, 5.0]
            )

            delta_step_g = np.full(shape=(5, 2), fill_value=5.0)
            delta_step_g[0, 0] = j / 10
            delta_step_g[0, 1] = i / 10

            matrix_z_vib_einzeln = calc_delta_intensities(
                phi,
                theta,
                trar1,
                trar2,
                Beam_variables,
                beam_indices,
                CDisp,
                E_array,
                VPI_array,
                VV_array,
                amplitudes_ref,
                amplitudes_del,
                n_files,
                NCSurf,
                delta_step_einzeln,
                number_z_steps,
            )

            matrix_z_vib_g = calc_delta_intensities_2D(
                phi_g,
                theta_g,
                trar1_g,
                trar2_g,
                Beam_variables_g,
                beam_indices_g,
                CDisp_g,
                E_array_g,
                VPI_array_g,
                VV_array_g,
                amplitudes_ref_g,
                amplitudes_del_g,
                n_files_g,
                NCSurf,
                delta_step_g,
                number_z_steps,
                number_vib_steps,
            )

            for e_index in range(len(E_array)):
                for b_index in range(int0):
                    matrix_differenz[e_index, b_index] = (
                        matrix_z_vib_g[e_index, b_index]
                        - matrix_z_vib_einzeln[e_index, b_index]
                    )

            Integral = 0
            for b in range(int0):
                Integral = Integral + abs(np.trapz(matrix_differenz[:, b]))

            array[j, i] = Integral

    return array


def PlotMaker1D(
    phi,
    theta,
    trar1,
    trar2,
    Beam_variables,
    beam_indices,
    CDisp,
    E_array,
    VPI_array,
    VV_array,
    amplitudes_ref,
    amplitudes_del,
    n_files,
    NCSurf,
    phi_g,
    theta_g,
    trar1_g,
    trar2_g,
    Beam_variables_g,
    beam_indices_g,
    CDisp_g,
    E_array_g,
    VPI_array_g,
    VV_array_g,
    amplitudes_ref_g,
    amplitudes_del_g,
    n_files_g,
    int0,
    number_z_steps,
):

    array4 = np.full(shape=(101), fill_value=np.NaN)
    matrix_differenz = np.full(shape=(len(E_array), int0), fill_value=np.NaN)

    for i in tqdm(range(101)):

        delta_step_einzeln = np.full(shape=(40), fill_value=5.0)
        delta_step_einzeln[0] = i * 1 / (10 * sqrt(2)) + 1.464
        delta_step_einzeln[1] = i * 1 / (10 * sqrt(2)) + 1.464

        delta_step_dig = np.full(shape=(20), fill_value=5.0)
        delta_step_dig[0] = i / 10

        matrix_z_vib_einzeln = calc_delta_intensities(
            phi,
            theta,
            trar1,
            trar2,
            Beam_variables,
            beam_indices,
            CDisp,
            E_array,
            VPI_array,
            VV_array,
            amplitudes_ref,
            amplitudes_del,
            n_files,
            NCSurf,
            delta_step_einzeln,
            number_z_steps,
        )

        matrix_z_vib_g = calc_delta_intensities(
            phi_g,
            theta_g,
            trar1_g,
            trar2_g,
            Beam_variables_g,
            beam_indices_g,
            CDisp_g,
            E_array_g,
            VPI_array_g,
            VV_array_g,
            amplitudes_ref_g,
            amplitudes_del_g,
            n_files_g,
            NCSurf,
            delta_step_dig,
            number_z_steps,
        )

        for e_index in range(len(E_array)):
            for b_index in range(int0):
                matrix_differenz[e_index, b_index] = (
                    matrix_z_vib_g[e_index, b_index]
                    - matrix_z_vib_einzeln[e_index, b_index]
                )

        Integral = 0
        for b in range(int0):
            Integral = Integral + abs(np.trapz(matrix_differenz[:, b]))

        array4[i] = Integral

    return array4



"""Module reader of viperleed.calc.files.parameters.

This module is based on the original version of readPARAMETERS from
@fkraushofer (2020). This module exists only to avoid cyclic imports
between .read and .write, as both need access to one of the classes
defined here (ParametersReader, RawLineParametersReader), as well as
some functionality from the other module.

Defines context-manager, iterator classes for reading parameters from
a PARAMETERS file.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-10-16'
__license__ = 'GPLv3+'

from collections.abc import Iterator
from contextlib import AbstractContextManager
from pathlib import Path
import re

from viperleed.calc.lib.base import strip_comments

from .errors import MissingEqualsError
from .errors import ParameterHasNoValueError
from .errors import ParameterNotRecognizedError
from .known_parameters import did_you_mean
from .known_parameters import from_alias
from .utils import Assignment


class ParametersReader(AbstractContextManager, Iterator):
    """A context manager that iterates the contents of a PARAMETERS file."""

    def __init__(self, filename, noisy=True):
        """Initialize instance.

        Parameters
        ----------
        filename : str or Path
            Path to the file to be read.
        noisy : bool, optional
            Whether the reader will emit logging messages and raise
            errors if unknown or malformed parameters are encountered.
        """
        self.filename = Path(filename)
        self.noisy = noisy
        self._file_obj = None
        self._current_line = 0

    def __enter__(self):
        """Enter context."""
        self._file_obj = self.filename.open('r', encoding='utf-8')
        self._current_line = 0
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Close file, then return control to caller to handle exceptions."""
        try:
            self._file_obj.close()
        except AttributeError:
            pass
        return super().__exit__(exc_type, exc_value, traceback)

    def __next__(self):
        """Return the next understandable information in the file."""
        for line in self._file_obj:
            self._current_line += 1
            param, *rest = self._read_one_line(line)
            if not param:
                continue
            return (param, *rest)
        raise StopIteration

    def _complain_about_line_parse_errors(self, line, exc):
        """Re-raise an exception occurred while line was being parsed."""
        if not isinstance(exc, ParameterNotRecognizedError):
            raise exc
        # Suggest a close match for an unknown parameter
        param, *_ = self._tokenize_line(line)
        try:
            close_match = did_you_mean(param)
        except ParameterNotRecognizedError:
            # No close match. Stick to the original exception
            raise exc from None
        new_message = (f'{exc.message} (at line {self._current_line}). '
                       f'Did you mean {close_match!r}?')
        raise ParameterNotRecognizedError(exc.parameter, message=new_message)

    def _complain_about_missing_equals(self, line):
        """Warn the user if line contains a known parameter but no '='."""
        if not line or not self.noisy:
            return
        param, *_ = line.split()
        try:
            param = from_alias(param)
        except ParameterNotRecognizedError:
            return
        _err_msg = (f'-- line {self._current_line} -- Found {param} '
                    'in a line without an "=" sign. Assignment will '
                    f'be SKIPPED.\n    Faulty line: {line}')
        raise MissingEqualsError(param, message=_err_msg)

    def _read_one_line(self, line):
        """Return a parameter and other custom information from one line."""
        line = strip_comments(line)

        # We treat STOP differently, as its presence, even without
        # an equal sign, is interpreted as 'please, stop right now'
        if self._stop_requested(line):
            return 'STOP', Assignment('True', 'STOP')

        if '=' not in line:
            self._complain_about_missing_equals(line)
            return '', None

        try:
            param, assignment = self._parse_line(line)
        except (ParameterNotRecognizedError, ParameterHasNoValueError) as exc:
            if not self.noisy:
                return '', None
            self._complain_about_line_parse_errors(line, exc)
        if not param:
            return '', None
        return param.upper(), assignment

    def _parse_line(self, line):
        """Return a parameter string and an Assignment from line."""
        param, flags, values_str = self._tokenize_line(line)
        if not param:
            return '', None
        param = from_alias(param)
        values_str = values_str.strip()
        if not values_str:
            raise ParameterHasNoValueError(parameter=param)
        assignment = Assignment(values_str=values_str,
                                parameter=param,
                                flags_str=flags)
        return param, assignment

    def _stop_requested(self, line):
        """Return whether line contains a STOP condition."""
        # SEARCH_KILL is legacy name
        line = line.upper()
        stop_false = {alias: re.compile(fr'\s*{alias}\s*=\s*[F](ALSE)?').match
                      for alias in ('SEARCH_KILL', 'STOP')}
        return any(line.startswith(stop) and not is_false(line)
                   for stop, is_false in stop_false.items())

    @staticmethod
    def _tokenize_line(line):
        """Split up line into tokens."""
        # Syntax is 'PARAMETER flag flag ... = value value value ...'
        left_side, values_str = line.split('=', maxsplit=1)
        if not left_side:  # Nothing left of '='
            return left_side, (), values_str
        param, *flags = left_side.split()
        return param, flags, values_str


# To me this pylint complaint does not make much sense
# here. The public methods come from the parent.
# pylint: disable-next=too-few-public-methods
class RawLineParametersReader(ParametersReader):
    """A ParametersReader that also returns lines exactly as they were read.

    Differently from a ParametersReader, a RawLineParametersReader
    returns all the lines, whether they contain a parameter or not.
    """

    def __next__(self):
        """Return the next acceptable information in the file."""
        self._current_line += 1
        return self._read_one_line(next(self._file_obj))

    def _read_one_line(self, line):
        """Return a parameter, and the whole raw line it was in."""
        param, *_ = super()._read_one_line(line)
        return param, line



"""Package special of viperleed.calc.classes.rparams.

This package defines classes for not-so-easy parameters to be used
as attributes for an Rparams object. These are, for example:
- parameters that would otherwise need to have variable types and
  require a bunch of isinstance checks when used
- parameters that have peculiar structure
- parameters that have multiple values that are cleaner when stored
  as attributes
- parameters that have a non-so-simple conversion to/from string
  or Assignment, i.e., those for which a ParameterInterpreter
  interpret_<...> method would be very complex.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-10-21'
__license__ = 'GPLv3+'



"""
========================================
   ViPErLEED Graphical User Interface
========================================
*** module guilib.leedsim.exportcsv ***

Created: 2021-01-21
Author: Michele Riva

This module contains functions needed for exporting the pattern files needed
as input for the ViPErLEED ImageJ plug-ins
"""

import numpy as np

from viperleed import guilib as gl
from viperleed import GLOBALS


def export_pattern_csv(fnames, leeds, **kwargs):
    """
    Exports the pattern file needed for the ViPErLEED ImageJ plug-in for
    extracting experimental I(V) curves.
    
    Parameters
    ----------
    - fnames: str, or list of str
        Paths to files to be saved. When passing multiple LEED patterns, if only
        one file name is given, the pattern files of each single structure are
        combined there. If multiple paths are given, the first one is used for
        the combined pattern file, all the others for the single-structure
        files. Thus in this case (n+1) paths need to be given, where
        n = len(leeds).
    - leeds: list of dict(s) or viperleed.LEEDPattern(s)
        List determining what has to be exported. All LEED patterns must have
        the same bulk basis. No attempt to make them consistent by a unitary
        transformation (rotate/mirror/swap unit vectors) is made.
    - keyword arguments:
        * version: str
            Current version of ViPErLEED that exported the pattern file. Will
            eventually be fetched from the GLOBALS once we decide on some
            versioning and how to store this info.
        * source: str
            Path to file (or any other useful reference to it) that was used
            as an input to generate the pattern file. In GUI, this is the .tlm
            file. When called from calc it may be related to the names given
            automatically by the bookkeeper.
        * name: str
            Descriptive text giving a name to the structure, independent of
            file name
        * domains: iterable of (iterable of int or None)
            the 'outer' list should contain as many entries as there are LEED
            parameters. Each entry is a list of indices (0-based) of the
            symmetry-equivalent domains to be exported. This is tricky to
            infer without looking at the GUI. If any of the lists is None, all
            symmetry-equivalent domains are passed. If None or not given, all
            symmetry-equivalent domains from all structures are exported.
    """
    # Check and process parameters:
    version = kwargs.get('version', GLOBALS['version'])
    source = kwargs.get('source', None)
    name = kwargs.get('name', None)
    all_domains = kwargs.get('domains', [None]*len(leeds))
    
    if isinstance(fnames, (list, tuple, np.ndarray)):
        # will export one combined file, plus as many files as
        # there are patterns
        fname_combined = fnames[0]
        fnames = fnames[1:] if fnames[1:] else [None]*len(leeds)
        if len(fnames) != len(leeds):
            raise ValueError("exportcsv: not enough file paths given for "
                             "exporting multiple LEED patterns")
    elif isinstance(fnames, str):
        fname_combined = fnames
        fnames = [None]*len(leeds)

    if len(all_domains) != len(leeds):
        raise ValueError("exportcsv: The number of domains definition indices "
                         f"({len(all_domains)}) is inconsistent with that of "
                         f"LEED patterns ({len(leeds)}).")
    # TODO: the whole block that follows will be done with a single
    #       call to the new LEEDPattern
    if any(not isinstance(leed, (dict, gl.LEEDPattern, gl.LEEDParameters))
           for leed in leeds):
        raise TypeError("exportcsv: each of the LEED parameters passed must "
                        "be either a leed_parameters dictionary or a "
                        "viperleed.LEEDPattern")
    for i, leed in enumerate(leeds):
        if isinstance(leed, (dict, gl.LEEDParameters)):
            # replace the dictionary entry with a LEEDPattern
            leeds[i] = gl.LEEDPattern(leed)
        # and check that the bulk bases are consistent
        if not np.allclose(leed.bulk_basis, leeds[0].bulk_basis):
            raise ValueError("exportcsv: Incompatible bulk bases found among "
                             "the LEED patterns")

    # TODO: Also this part needs to change. I will do the loop below
    #       only if there are fnames to save, iterating through the
    #       leed.domains
    #       In all cases, I'll then call _format_beams_ on the full LEEDPattern
    #       to get the combination of all structures at once

    # Now go for the actual code
    all_lines = []
    for leed, domains, fname in zip(leeds, all_domains, fnames):               # with the new LEEDPattern, leeds -> leed.domains, so that each iteration brings out a LEEDSymmetryDomains
        lines = _format_beams_(leed,
                               domains=domains,
                               name=name,
                               version=version,
                               source=source)
        if fname:  # write to file the single patterns
            with open(fname, 'w+') as f:
                f.write('\n'.join(lines))
        all_lines.append(lines)
    
    # and export the combined one
    combined = _combine_pattern_files_(all_lines)
    with open(fname_combined, 'w+') as f:
        f.write('\n'.join(combined))


def _combine_pattern_files_(all_lines):
    """
    TO BE IMPLEMENTED
    """
    return all_lines[0]


def _format_beams_(leed, **kwargs):
    """
    Format beams BLAH BLAH
    
    Parameters
    ----------
    - leed: guilib.LEEDPattern
    - keyword arguments:
        * domains: iterable of int
            Indices (0-based) of the domains to be exported. This is tricky to
            infer without looking at the GUI
        * others are passed to _format_header_
    """

    # Check parameters
    if not isinstance(leed, gl.LEEDPattern):                                   # This will not be a LEEDPattern, but rather a LEEDSymmetryDomains
        raise TypeError("exportcsv: leed must be a "
                        "viperleed.guilib.LEEDPattern instance")

    beams = leed.get_equivalentSpots(kwargs.get('domains', None))
    fractions, groups, overlaps, extinct_domains = zip(*beams)

    # Process the data to get them in a better format for packing each
    # line; also find the correct number of characters for each column
    hk = []  # beam indices
    gg = []  # reciprocal lattice vectors
    dd = []  # overlapping domains

    # lengths is a dictionary that will contain the maximum
    # number of characters appearing in each part of the columns
    lengths = {'numerator': [], 'denominator': [], 'hk_integer': [],
               'g_integer': [], 'domains': []}

    for fract in fractions:
        # construct beam (= hk indices) and g (= reciprocal-space vector
        #                                      in Cartesian coordinates)
        beam = gl.BeamIndex(fract)
        hk.append(beam)
        g = np.dot(beam, leed.bulk_basis)
        gg.append(g)

        # get relevant lengths:
        # 1) fractional representation of beam indices
        num, den = beam.get_format_lengths('s')
        lengths['numerator'].append(num)
        lengths['denominator'].append(den)

        # 2) float representation of beam indices
        lengths['hk_integer'].append(beam.get_format_lengths('f')[0])

        # 3) and reciprocal-space vector
        lengths['g_integer'].append(gl.integer_part_length(*g))

    # construct the list of overlapping domains
    for (overlap, extinct) in zip(overlaps, extinct_domains):
        overlap_txt = [str(dom) for dom in overlap]
        for e, (s, ext) in enumerate(zip(overlap_txt, extinct)):
            if ext:
                overlap_txt[e] = f"({s})"
        overlap_txt = '+'.join(dom for dom in overlap_txt)
        dd.append(overlap_txt)
        lengths['domains'].append(len(overlap_txt))

    # get the maximum of all lengths, which will determine the formatting
    for key in lengths.keys():
        lengths[key] = max(lengths[key])

    # and account also for the text in the column headers
    lengths['group'] = max(len(str(np.max(groups))), len('group'))
    lengths['domains'] = max(lengths['domains'], len('domain(s)'))

    to_export = [*_format_header_(lengths, leed, **kwargs)]

    for (beam, g, group, doms) in zip(hk, gg, groups, dd):
        n, d = lengths['numerator'], lengths['denominator']
        
        line = (    # list of entries for each column:
            f"{beam:({n},{d})s},"                              # fractional hk
            + f"{beam:{lengths['hk_integer']}f}"[1:-1] + ','   # floating hk
            + gl.format_floats(f"{lengths['g_integer']}f", *g) + ','  # gx, gy
            + f"{group:>{lengths['group']}},"                  # group index
            + f"{doms:>{lengths['domains']}},"                 # domains overlap
            )

        to_export.append(line)

    return to_export


def _format_header_(lengths, leed, **kwargs):                                   # TODO: This needs to handle both leed=LEEDStructuralDomains and leed=LEEDSymmetryDomains
    """
    This function is used internally by export_pattern_file.

    Parameters
    ----------
    - lengths: dict
        Dictionary, generated by _format_beams_ with the following mandatory
        keys, used for formatting the column headers:
        * 'numerator'
        * 'denominator'
        * 'hk_integer'
        * 'g_integer'
        * 'group'
        * 'domains'
    - leed: guilib.LEEDPattern
        A full LEEDPattern that is used for exporting.
    - keyword arguments:
        * version: str
            Current version of ViPErLEED that exported the pattern file. Will
            eventually be fetched from the GLOBALS once we decide on some
            versioning and how to store this info.
        * source: str
            Path to file (or any other useful reference to it) that was used
            as an input to generate the pattern file. In GUI, this is the .tlm
            file. When called from calc it may be related to the names given
            automatically by the bookkeeper.
        * name: str
            Descriptive text giving a name to the structure, independent of
            file name
        * domains: iterable of int
            Indices (0-based) of the domains to be exported. This is tricky to
            infer without looking at the GUI
    Returns
    -------
    - list of str: list of lines to be exported
    
    """
    # 1st Header
    # * ViPErLEED header
    # * reference to source used for exporting (path or other)
    #
    # 2nd Header:
    # * optional structure name
    # * max energy
    # * bulk shape (group); lattice parameters and basis
    # * surf shape (group); lattice parameters
    # * total number of domains, and number of domains exported
    # * for each exported domain: basis and superlattice matrix
    # * finally an uncommented header line for the columns

    # check mandatory parameters
    if not isinstance(leed, gl.LEEDPattern):
        raise TypeError("exportcsv: leed must be a "
                        "viperleed.guilib.LEEDPattern instance")
    needed_keys = ('numerator', 'denominator', 'hk_integer', 'g_integer',
                   'group', 'domains')
    if any((key not in lengths.keys()) for key in needed_keys):
        raise ValueError("exportcsv: some keys are missing from the required "
                         "parameter 'lengths'.")

    # Parse kwargs:
    version = kwargs.get('version', GLOBALS['version'])
    source = kwargs.get('source', None)
    name = kwargs.get('name', None)
    export_domains = kwargs.get('domains', range(leed.n_domains))

    # Fill up the header
    header = ["# This file was automatically generated by ViPErLEED"
              + f" v{version}"]
    if source:
        header.append("# It contains the pattern file generated from:\n# "
                      + source)
        header.append('#')
    if name:
        header.append(f"# Structure: {name}\n#")

    header.append(f"# Max. LEED Energy: {leed.max_energy:.1f} eV")
    
    # TODO: add incidence angles

    # TODO: the next block inevitably needs to change, as we have to be
    #       more general to handle multiple structures, so bulk and surface(s)
    #       must be treated separately.
    #       probably easiest: (1) when a LEEDSymmetryDomains is passed, just
    #       wrap it into a 1-element tuple. (2) Define an extra indent if there
    #       is more than one pattern to process. (3) The line "Surface
    #       lattice..." will need to be different if there is a single
    #       structure or multiple ones. Probably something like "Structural
    #       domain {name}: ..." in the latter case
    
    #       LEEDSymmetryDomains has a .bulk property (a Lattice), and self[0]
    #       is the first-domain Lattice

    for txt in ('Bulk', 'Surface'):
        lattice = leed.reciprocal_lattices[txt[:4].lower()]
        basis = lattice.reciprocal_basis  # TODO: probably better to use real_basis
        shape = lattice.cell_shape
        group = lattice.group.group
        basis_text = _format_lattice_basis_(basis, shape)
        header.append(f"# {txt} lattice: {shape} ({group}); {basis_text}")
        if txt == 'Bulk':
            header.append('#' + ' '*8 + _format_basis_vectors_(basis))

    basis = leed.reciprocal_lattices['bulk'].real_basis

    # TODO: the next part will just be a loop over the structures, using
    #       the extra indent from above
    matrices = leed.superlattices[export_domains]

    header.append(f"#\n# Exporting beams from {len(matrices)} domain(s) "
                  f"out of {leed.n_domains} symmetry-equivalent ones")

    for (dom, m) in zip(export_domains, matrices):
        super_basis = np.dot(m, basis)
        dom_txt = f" Domain {dom + 1} - "
        header.append(f"#{dom_txt}{_format_basis_vectors_(super_basis)}")
        header.append('#' + ' '*len(dom_txt) + 'Superlattice: '
                      + np.array2string(m, separator=' ').replace('\n',''))

    # Prepare the headers of the columns
    # 0) Some description
    header.append(
        '#\n'
        '# * h and k are the surface Miller indices of each LEED spot;'
        ' both fractional\n'
        '#   and floating-point versions are provided.\n'
        '# * gx and gy are the horizontal and vertical components of'
        ' reciprocal lattice\n'
        '#   vectors in AA^(-1), and include a factor of 2*pi.\n'
        '# * Beams with the same absolute value of "group" are'
        ' symmetry equivalent (at\n'
        '#   normal incidence); extinct spots have a negative "group"'
        ' index.\n'
        '# * The domains contributing to each spot are listed in'
        ' "domain(s)". Domains\n'
        '#   contributing with glide-extinct spots are reported in'
        ' parentheses.\n'
        )

    # then all column headers centered with the contents
    n, d = lengths['numerator'], lengths['denominator']
    fract_width = n + d

    # if the indices are fractions, add 1 to account for the "/"
    fract_width += 1 if lengths['denominator'] else 0

    column_headers = (
        f"({'h':^{fract_width}}|{'k':^{fract_width}}),"  # fractional (h|k)
        f"{'h':^{lengths['hk_integer'] + 6}},"           # float h
        f"{'k':^{lengths['hk_integer'] + 6}},"           # float k
        f"{'gx':^{lengths['g_integer'] + 6}},"           # float gx
        f"{'gy':^{lengths['g_integer'] + 6}},"           # float gy
        f"{'group':^{lengths['group']}},"                # group
        f"{'domain(s)':^{lengths['domains']}},"          # domains
        )

    header.append(column_headers)

    return header


def _format_lattice_basis_(basis, shape):
    a, b = np.linalg.norm(basis, axis=1)
    alpha = np.degrees(np.arccos(np.dot(basis[0], basis[1])/(a*b)))

    txt = [f"a = {a:.4f} AA"]

    if shape not in ['Square', 'Hexagonal', 'Rhombic']:
        txt.append(f"b = {b:.4f} AA")
    if shape not in ['Square', 'Rectangular']:
        txt.append(f"alpha = {alpha:.2f} deg")

    return '; '.join(txt)


def _format_basis_vectors_(basis):
    txt = (
        'Basis: '
        + 'a = {}'.format(np.array2string(basis[0], precision=4,
                                          floatmode='fixed', separator=' ',
                                          suppress_small=True))
        + '; '
        + 'b = {}'.format(np.array2string(basis[1], precision=4,
                                          floatmode='fixed', separator=' ',
                                          suppress_small=True))
        )
    return txt




"""Section Superpos."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Michael Riva (@michele-riva)'
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-11'
__license__ = 'GPLv3+'

import copy
import logging
import os
from pathlib import Path
import shutil
import subprocess

from viperleed.calc.files import iosuperpos
from viperleed.calc.files.beams import averageBeams
from viperleed.calc.files.beams import writeFdOut
from viperleed.calc.files.beams import writeOUTBEAMS
from viperleed.calc.files.displacements import readDISPLACEMENTS_block
from viperleed.calc.files.iorefcalc import readFdOut
from viperleed.calc.files.iosearch import readSDTL_blocks
from viperleed.calc.files.iosearch import readSDTL_end
from viperleed.calc.lib import leedbase
from viperleed.calc.lib.checksums import validate_multiple_files

logger = logging.getLogger(__name__)


def superpos(sl, rp, subdomain=False, for_error=False, only_vary=None):
    """Runs the superpos calculation."""
    # check whether there is anything to evaluate
    if for_error:
        config = ((None, None),)
    elif not for_error and rp.searchResultConfig is not None:
        config = rp.searchResultConfig[0]
        only_vary = None  # Meant to use only for_error
    else:  # not for_error, but we don't have the search results handy
        config = _best_config_from_sdtl(rp)
        only_vary = None  # Meant to use only for_error

    if not config:
        # Errors in _best_config_from_sdtl, reported already
        return

    if rp.domainParams:
        superpos_domains(rp, config)
        return

    # read DISPLACEMENTS block and fetch deltas
    if not rp.disp_block_read:
        readDISPLACEMENTS_block(rp, sl, rp.disp_blocks[rp.search_index])
        rp.disp_block_read = True
    if not any(ind in rp.runHistory for ind in [2, 3]) and not for_error:
        leedbase.getDeltas(rp.TENSOR_INDEX, required=True)
    # make sure search parameters are initialized
    if 3 not in rp.runHistory and not subdomain and not for_error:
        logger.debug("Superpos calculation executed without search. "
                     "Search parameters will be inferred from input files.")
        try:
            rp.generateSearchPars(sl)
        except Exception:
            logger.error("Error getting search parameters. Superpos will "
                         "stop.")
            return

    # now we have configuration and parameters, create input:
    contrin = ""
    try:
        contrin = iosuperpos.writeSuperposInput(sl, rp, config[0][1],
                                                for_error=for_error,
                                                only_vary=only_vary)
    except Exception:
        logger.error("Error getting input data for Superpos: ", exc_info=True)
        rp.setHaltingLevel(2)
        return

    if not contrin:
        logger.error("Error getting input data for Superpos: "
                     "writeSuperposInput returned empty. Cancelling "
                     "Superpos...")
        rp.setHaltingLevel(2)
        return

    logger.debug("Wrote Superpos input successfully")
    # if execution is suppressed, stop here:
    if rp.SUPPRESS_EXECUTION:
        logger.warning("SUPPRESS_EXECUTION parameter is on. Superpos "
                       "calculation will not proceed. Stopping...")
        rp.setHaltingLevel(3)
        return

    if not rp.FORTRAN_COMP[0]:
        rp.getFortranComp()
    # Get FORTRAN files                                                         # TODO: use CompileTask subclass (Issue #43)
    try:
        tl_source = rp.get_tenserleed_directory()
        tl_path = tl_source.path
        src_dir = tl_path / 'src'
        src_path = next(f for f in src_dir.glob('superpos*'))                   # TODO: StopIteration; Probably also want *.f* in the glob?
        shutil.copy2(src_path, src_path.name)
        lib_dir = tl_path / 'lib'
        lib_path = next(f for f in lib_dir.glob('lib.superpos*'))               # TODO: StopIteration; Probably also want *.f* in the glob?
        shutil.copy2(lib_path, lib_path.name)
        globalname = "GLOBAL"
        shutil.copy2(src_dir / globalname, globalname)
    except Exception:
        logger.error("Error getting TensErLEED files for superpos: ")
        raise

    # Validate checksums
    if not rp.TL_IGNORE_CHECKSUM:
        files_to_check = (lib_path,
                          src_path,
                          src_dir / globalname)
        validate_multiple_files(files_to_check, logger,
                                "superpos", rp.TL_VERSION)

    # Compile FORTRAN files
    sposname = Path(f"superpos-{rp.timestamp}")
    logger.info("Compiling fortran input files...")

    compile_log = "compile-superpos.log"
    try:
        leedbase.fortran_compile(f"{rp.FORTRAN_COMP[0]} -o",
                                 f"{sposname} {src_path.name} {lib_path.name}",
                                 rp.FORTRAN_COMP[1],
                                 logname=compile_log)
    except Exception:
        leedbase.copy_compile_log(rp, Path(compile_log), "superpos-compile")
        logger.error("Error compiling fortran files: ", exc_info=True)
        raise

    logger.info("Starting Superpos calculation...")
    outname = "superpos-spec.out"
    err_log = ""
    try:
        with open(outname, "w") as out:
            complete = subprocess.run(sposname.resolve(),                       # TODO: Perhaps nicer to have log be a TextIO stream?
                                      input=contrin, encoding="ascii",
                                      capture_output=True)
            out.write(complete.stdout)
            err_log = complete.stderr
    except Exception:
        logger.error("Error during Superpos calculation.")
        raise
    err_log = "\n".join(line for line in err_log.splitlines()
                        if ".  CORRECT TERMINATION" not in line)
    if err_log:
        logger.warning("Superpos output contained the following "
                       f"warnings/error messages:\n{err_log}")
    logger.info("Finished Superpos calculation. Processing files...")
    try:
        beams_and_spec = readFdOut(outname, for_error=for_error)
        # if for_error, theobeams will contain only the first variation
    except FileNotFoundError:
        logger.error(f"{outname} not found after superpos calculation.")
        raise
    except Exception:
        logger.error(f"Error reading {outname} after superpos calculation.")
        raise
    rp.theobeams["superpos"], rp.superpos_specout = beams_and_spec

    if not for_error:
        _write_fitbeams(rp)

    # Rename and move files
    try:
        os.rename('PARAM', 'superpos-PARAM')
    except Exception:
        logger.warning("Failed to rename superpos input file PARAM to "
                       "superpos-PARAM")
    try:
        os.rename('DOC', 'superpos-DOC')
    except Exception:
        pass


def superpos_domains(rp, configs):
    """Runs the normal superpos function for each subdomain, collects the
    results and averages over the beams to get an overall result."""
    # make sure search parameters are initialized
    if 3 not in rp.runHistory:
        logger.debug("Superpos calculation executed without search. "
                     "Search parameters will be inferred from input files.")
        try:
            rp.generateSearchPars(None)
        except Exception:
            logger.error("Error getting search parameters. Superpos will "
                         "stop.")
            return
    home = os.getcwd()
    percentages = []
    for (percent, params), dp in zip(configs, rp.domainParams):
        percentages.append(percent)
        dp.rp.searchResultConfig = [[(100, params)]]
        logger.info(f"Running superpos calculation for domain {dp.name}")
        try:
            os.chdir(dp.workdir)
            superpos(dp.sl, dp.rp, subdomain=True)
        except Exception:
            logger.error("Error while running superpos "
                         f"calculation for domain {dp.name}")
            raise
        finally:
            os.chdir(home)

    logger.info("Getting weighted average over domain beams...")
    rp.theobeams["superpos"] = averageBeams(
        [dp.rp.theobeams["superpos"] for dp in rp.domainParams],
        weights=percentages
        )
    _write_fitbeams(rp)

    try:
        rp.superpos_specout = writeFdOut(rp.theobeams["superpos"], rp.beamlist,
                                         filename="superpos-spec.out",
                                         header=rp.systemName)
    except Exception:
        logger.error("Error writing averaged superpos-spec.out for R-factor "
                     "calculation.", exc_info=rp.is_debug_mode)


def _best_config_from_sdtl(rp):                                                 # TODO: more doc
    """Return the best configuration from the most recent search."""
    # check for an SD.TL file
    cwd = Path()
    sdtl_files = (f for f in (cwd/"SD.TL", cwd/"OUT"/"SD.TL") if f.is_file())
    sdtl_file = next(sdtl_files, None)
    if not sdtl_file:
        logger.error("Superpos: Found no stored results from recent "
                     "search and no SD.TL file. Cancelling...")
        rp.setHaltingLevel(2)
        return None

    try:
        sdtl = readSDTL_end(filename=str(sdtl_file),
                            n_expect=rp.SEARCH_POPULATION)
    except Exception:                                                           # TODO: better exception
        logger.error("Superpos: Error reading SD.TL")
        rp.setHaltingLevel(2)
        return None

    if sdtl is not None:
        sdtl = readSDTL_blocks("\n".join(sdtl),
                               whichR=rp.SEARCH_BEAMS,
                               n_expect=rp.SEARCH_POPULATION)
    if not sdtl:
        logger.error("Superpos: No data found in SD.TL")
        rp.setHaltingLevel(2)
        return None

    try:
        return sdtl[0][2][0]  # Most recent block, config, best only
    except IndexError:
        logger.error("Superpos: Failed to read best configuration from "
                     "SD.TL")
        rp.setHaltingLevel(2)
        return None


def _write_fitbeams(rp):                                                        # TODO: Could probably be made more general by passing in which rp beams, a base filename, and a Section (or even rp and a Section)
    """Write FITBEAMS and its _norm version."""
    try:
        writeOUTBEAMS(rp.theobeams["superpos"], filename="FITBEAMS.csv")
        theobeams_norm = copy.deepcopy(rp.theobeams["superpos"])
        for beams in theobeams_norm:
            beams.normMax()
        writeOUTBEAMS(theobeams_norm, filename="FITBEAMS_norm.csv")
    except Exception:
        logger.error("Error writing FITBEAMS after superpos calculation.",
                     exc_info=rp.is_debug_mode)



"""Module layer of viperleed.calc.classes.

Classes storing position and atom list of a Layer, and its SubLayer
subclass. The latter is for atoms of the same chemical species at
the same z position.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2019-06-13'
__license__ = 'GPLv3+'

from viperleed.calc.classes.atom_containers import AtomContainer
from viperleed.calc.classes.atom_containers import AtomContainerError


class LayerError(AtomContainerError):
    """Base exception for Layer objects."""


class LayerHasNoAtomsError(LayerError):
    """Operation cannot be performed as the layer is empty."""


class Layer(AtomContainer):                                                     # TODO: modify description of .cartori when flipping .cartpos[2] -- Issue #174
    """A container of atoms residing close to one another along z.

    This is intended to be used with Slab objects. Has origin, atoms
    (a subset of the ones in slab), and a number.

    Attributes
    ----------
    atlist : list of Atom
        The atoms that belong to this layer.
    cartori : numpy.ndarray
        The Cartesian origin of this layer. It is the position where a
        plane passing through the topmost atom in this layer intersects
        the c vector of the slab's unit cell. Notice that `cartori` is
        in the same Cartesian frame as the atoms (i.e., z increases
        going from vacuum into the solid).
        A call to update_position() updates this attribute.
    cartbotz : float
        Z (i.e., out-of-plane) position of the bottom-most atom in
        the layer. A call to update_position() updates this attribute.
    is_bulk : bool
        Whether this layer has bulk character.
    num : int
        A progressive index (zero-based) identifying this layer
        within its slab. Normally, `layer.num == 0` for the layer
        closest to the solid/vacuum interface.
    slab : Slab
        The slab to which this layer belongs.
    """

    def __init__(self, slab, num, is_bulk=False):
        """Initialize instance."""
        self.slab = slab
        self.num = num
        self.is_bulk = is_bulk
        self.atlist = []

        # Note about the next two: they should be kept in memory,
        # rather than re-determined on-the-fly, since the position
        # of the atoms in question may change during a search, but
        # TensErLEED expects atom positions to be referred to the
        # layer position as it was used in the reference calculation
        self.cartori = None
        self.cartbotz = None

    def __contains__(self, value):
        """Return whether this layer contains an atom."""
        return value in self.atlist

    def __iter__(self):
        """Return an iterator of Atoms in this Layer."""
        return iter(self.atlist)

    @property
    def n_atoms(self):
        """Return the number of atoms in this layer."""
        return len(self.atlist)

    @property
    def thickness(self):
        """Return the z thickness of this layer.

        Note
        ----
        This is calculated from the stored positions of this layer,
        and may not be up to date if atoms in this layer have moved
        along z since the last call to update_position().
        """
        return abs(self.cartbotz - self.cartori[2])

    def update_position(self):
        """Update the Cartesian position of this layer from its atoms."""
        if not self.atlist:
            raise LayerHasNoAtomsError(
                f'{type(self).__name__} needs atoms to update_position()'
                )
        sorted_atoms = sorted(self, key=lambda atom: atom.pos[2])
        topat = sorted_atoms[-1]
        botat = sorted_atoms[0]

        self.cartbotz = botat.cartpos[2]
        self.cartori = topat.pos[2] * self.slab.c_vector
        # So far x and y are correct, but z is still in the wrong               # TODO: this will not be necessary when we flip .cartpos[2] -- Issue #174
        # direction and with origin as POSCAR. Take the z directly
        # from the highest atom
        self.cartori[2] = topat.cartpos[2]


class SubLayer(Layer):                                                          # TODO: modify description of .cartori when flipping .cartpos[2] -- Issue #174
    """A Layer with the same chemical element and the same z position.

    Attributes
    ----------
    atlist : list of Atom
        The atoms that belong to this sublayer.
    cartori : numpy.ndarray
        The Cartesian origin of this layer. It is the position where a
        plane passing through the topmost atom in this layer intersects
        the c vector of the slab's unit cell. Notice that `cartori` is
        in the same Cartesian frame as the atoms (i.e., z increases
        going from vacuum into the solid).
        A call to update_position() updates this attribute.
    cartbotz : float
        Z (i.e., out-of-plane) position of the bottom-most atom in
        the layer. A call to update_position() updates this attribute.
    cartpos : numpy.ndarray
        The Cartesian position of this sublayer. Notice that this is
        not an alias for `cartori`. The (x, y) components of the two
        are not the same: `cartori[:2]` is a point along the c vector
        of the unit cell, while `cartpos[:2]` may be a point inside
        (or outside if atoms are not collapsed) the unit cell.
    element : str
        The chemical element of the atoms of this sublayer.
    is_bulk : bool
        Whether this layer has bulk character.
    num : int
        A progressive index (zero-based) identifying this sublayer
        within its slab. Normally, `sublayer.num == 0` for the
        sublayer closest to the solid/vacuum interface.
    pos : numpy.ndarray
        The fractional position of this sublayer. This is the
        fractional position corresponding to `cartpos`, not the
        one corresponding to `cartori`.
    slab : Slab
        The slab to which this sublayer belongs.
    symposlist : list
        Candidate positions for rotation / mirror / glide planes.
    """

    def __init__(self, slab, num, is_bulk=False):
        """Initialize instance."""
        super().__init__(slab, num, is_bulk=is_bulk)
        self.symposlist = []

    def __str__(self):
        """Return a string version of this sublayer."""
        n_atoms = self.n_atoms
        if not n_atoms:
            atoms = '(empty)'
        elif n_atoms == 1:
            atoms = self.atlist[0]
        else:
            atoms = [f'{at}' for at in self]
        return f'Sublayer(num={self.num}, {atoms}, cartbotz={self.cartbotz})'

    @property
    def element(self):
        """Return the chemical element for this sublayer."""
        if not self.atlist:
            raise LayerHasNoAtomsError(
                f'A {type(self).__name__} without atoms has no element'
                )
        return self.atlist[0].el

    @property
    def cartpos(self):
        """Return the Cartesian position of this sublayer."""
        if not self.atlist:
            raise LayerHasNoAtomsError(
                f'A {type(self).__name__} without atoms has no cartpos'
                )
        return self.atlist[0].cartpos

    @property
    def pos(self):
        """Return the fractional position of this sublayer."""
        if not self.atlist:
            raise LayerHasNoAtomsError(
                f'A {type(self).__name__} without atoms has no pos'
                )
        return self.atlist[0].pos



"""Module known_parameters of viperleed.calc.files.parameters.

Initial version by @fkraushofer in 2020, major rewrite by @amimre
and @michele-riva in June 2023. This module used to be part of
parameters.py. Refactored in October 2023.

Contains definitions of all known parameters and their supported
aliases. It also defines useful functions for retrieving parameter
names used internally when reading/writing/interpreting a PARAMETERS
file.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-18'
__license__ = 'GPLv3+'

from difflib import get_close_matches
import logging

from viperleed.calc.lib.base import parent_name

from .errors import ParameterNotRecognizedError


_LOGGER = logging.getLogger(parent_name(__name__))


# Allowed parameters with their 'standard' names in alphabetic order
KNOWN_PARAMS = (
    'ATTENUATION_EPS',
    'AVERAGE_BEAMS',
    'BEAM_INCIDENCE',
    'BULKDOUBLING_EPS',
    'BULKDOUBLING_MAX',
    'BULK_LIKE_BELOW',
    'BULK_REPEAT',
    'DOMAIN',
    'DOMAIN_STEP',
    'ELEMENT_MIX',
    'ELEMENT_RENAME',
    'FILAMENT_WF',
    'FORTRAN_COMP',
    'HALTING',
    'INTPOL_DEG',
    'IV_SHIFT_RANGE',
    'LAYER_CUTS',
    'LAYER_STACK_VERTICAL',
    'LMAX',
    'LOG_LEVEL',
    'LOG_SEARCH',
    'N_BULK_LAYERS',
    'N_CORES',
    'OPTIMIZE',
    'PARABOLA_FIT',
    'PHASESHIFT_EPS',
    'PHASESHIFTS_CALC_OLD',
    'PHASESHIFTS_OUT_OLD',
    'PLOT_IV',
    'RUN',
    'R_FACTOR_LEGACY',
    'R_FACTOR_SMOOTH',
    'R_FACTOR_TYPE',
    'SCREEN_APERTURE',
    'SEARCH_BEAMS',
    'SEARCH_CONVERGENCE',
    'SEARCH_CULL',
    'SEARCH_MAX_GEN',
    'SEARCH_POPULATION',
    'SEARCH_START',
    'SITE_DEF',
    'SUPERLATTICE',
    'SUPPRESS_EXECUTION',
    'STOP',
    'SYMMETRIZE_INPUT',
    'SYMMETRY_BULK',
    'SYMMETRY_CELL_TRANSFORM',
    'SYMMETRY_EPS',
    'SYMMETRY_FIND_ORI',
    'SYMMETRY_FIX',
    'TENSOR_INDEX',
    'TENSOR_OUTPUT',
    'THEO_ENERGIES',
    'TL_VERSION',
    'TL_IGNORE_CHECKSUM',
    'T_DEBYE',
    'T_EXPERIMENT',
    'V0_IMAG',
    'V0_REAL',
    'V0_Z_ONSET',
    'VIBR_AMP_SCALE',
    'ZIP_COMPRESSION_LEVEL',
    )


_DEPRECATED = {  # alias: (since version, till version), both inclusive
    'parabolafit': ('0.11.0', None),
}


def _make_alias(param):
    """Return an alias of param."""
    return param.lower().replace('_', '')


# _PARAM_ALIAS keys should be all lowercase, with no underscores
_PARAM_ALIAS = {_make_alias(p): p for p in KNOWN_PARAMS}
_PARAM_ALIAS.update({    # Sort keys alphabetically!
    'bulklike': 'BULK_LIKE_BELOW',
    'bulksymmetry': 'SYMMETRY_BULK',
    'compiler': 'FORTRAN_COMP',
    'compression': 'ZIP_COMPRESSION_LEVEL',
    'compression_level': 'ZIP_COMPRESSION_LEVEL',
    'fdoptimize': 'OPTIMIZE',
    'fdoptimization': 'OPTIMIZE',
    'fortrancompile': 'FORTRAN_COMP',
    'fortrancompiler': 'FORTRAN_COMP',
    'ignorechecksum': 'TL_IGNORE_CHECKSUM',
    'ivplot': 'PLOT_IV',
    'logdebug' : 'LOG_LEVEL',
    'plotrfactor': 'PLOT_IV',
    'plotrfactors': 'PLOT_IV',
    'searchkill': 'STOP',
    })


def from_alias(known_param_or_alias):
    """Return a parameter name, possibly from its alias.

    Parameters
    ----------
    known_param_or_alias : str
        The name of a parameter or one of its aliases.

    Returns
    -------
    known_param : str
        The name of the known parameter.

    Raises
    ------
    TypeError
        If known_param_or_alias is not a string.
    ParameterNotRecognizedError
        If known_param_or_alias is not a known parameter.
    """
    if known_param_or_alias in KNOWN_PARAMS:
        return known_param_or_alias
    try:
        return _PARAM_ALIAS[_make_alias(known_param_or_alias)]
    except KeyError:
        pass
    except AttributeError as exc:  # No .lower or no .replace
        raise TypeError('from_alias: must be string') from exc
    raise ParameterNotRecognizedError(parameter=known_param_or_alias)


def did_you_mean(unknown_param):
    """Return one close match for `unknown_param` from the known ones."""
    try:
        alias = _make_alias(unknown_param)
    except AttributeError as exc:  # No .lower or no .replace
        raise TypeError('from_alias: must be string') from exc
    close_matches = get_close_matches(alias, _PARAM_ALIAS)
    if not close_matches:
        raise ParameterNotRecognizedError(unknown_param)
    return _PARAM_ALIAS[close_matches[0]]


def is_deprecated(param, version):
    """Return whether param is deprecated for a ViPErLEED version."""
    try:
        since, till = _DEPRECATED.get(_make_alias(param))
    except (TypeError, KeyError):
        return False
    if version < since:
        return False
    if till is not None and version > till:
        return False
    return True


def warn_if_deprecated(param, version):
    """Emit a log warning if `param` is deprecated for `version`.

    Parameters
    ----------
    param : str
        Parameter to be checked for deprecation. It may also be
        one of the aliases.
    version : str
        The ViPErLEED version for which deprecation should be
        checked.

    Returns
    -------
    is_deprecated : bool
        True if `param` is deprecated for `version`.
    """
    if not is_deprecated(param, version):
        return False
    _LOGGER.warning(f'PARAMETER {param} is deprecated in ViPErLEED version '
                    f'{version}. This may cause errors or undefined behavior. '
                    'Use at your own risk')
    return True
